"""
MIT License

Copyright (c) 2017 cgalleguillosm

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from os import makedirs as _makedir, listdir as _listdir
from os.path import isfile as _isfile, isdir as _isdir, join as _join
from ntpath import split as _split, basename as _basename
from shutil import rmtree as _rmtree
from json import dump as _dump, load as _load


def file_exists(file_path, boolean=False, raise_error=False, head_message=''):
    """

    :param file_path:
    :param boolean:
    :param raise_error:
    :param head_message:
    :return:
    """
    exists = _isfile(file_path)
    if (raise_error or head_message != '') and not exists:
        raise Exception('{}{} File does not exist.'.format(head_message, file_path))
    if boolean:
        return exists
    return file_path


def dir_exists(dir_path, create=False):
    """

    :param dir_path:
    :param create:
    :return:
    """
    exists = _isdir(dir_path)
    if create and not exists:
        _makedir(dir_path)
        return True
    return exists


def path_leaf(path):
    """

    :param path:
    :return:
    """
    """

    Extract path and filename

    :param path: Entire filepath

    :return: Return a tuple that contains the (path, filename)

    """
    head, tail = _split(path)
    return (head, tail or _basename(head))


def remove_dir(path, parent_folder='results/'):
    """

    :param path:
    :param parent_folder:
    :return:
    """
    if path.startswith(parent_folder) and dir_exists(path):
        _rmtree(path)


def find_file_by(_path, prefix=None, sufix=None):
    """

    :param _path:
    :param prefix:
    :param sufix:
    :return:
    """
    filename = None
    for _filename in _listdir(_path):
        _filepath = _join(_path, _filename)
        if _isfile(_filepath) and (_filename.startswith(prefix) if prefix else _filename.endswith(sufix)):
            if not filename:
                filename = _filename
            else:
                raise Exception('Multiple benchmark files in a same folder.')
    return filename


def save_jsonfile(filepath, _dict, **kwargs):
    """

    :param filepath:
    :param _dict:
    :param kwargs:
    :return:
    """
    """
    
    Saves a json file.
    
    :param config_fp: Filepath to the config
    :param \*\*kwargs: Source for the config data  
    
    """
    with open(filepath, 'w') as file:
        _dump(_dict, file, **kwargs)


def load_jsonfile(filepath):
    """

    :param filepath:
    :return:
    """
    """
    
    Loads an specific json file
    
    :param filepath: Filepath of a json file.
    
    :return: A dictionary 
    
    """
    with open(filepath) as file:
        return _load(file)

class PlainFileReader:
    
    def __init__(self, filepath):
        """

        :param filepath:
        """
        assert(file_exists(filepath, True, True))
        self.file = open(filepath)
        self.last_pos = self.file.tell()
        self.EOF = False
        
    def nextline(self):
        """

        :return:
        """
        line = self.file.readline()
        last_pos = self.file.tell()
        if self.last_pos == last_pos:
            self.EOF = True
            return None
        self.last_pos = last_pos
        return line


"""
MIT License

Copyright (c) 2017 cgalleguillosm, AlessioNetti

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import logging

from inspect import stack
from time import perf_counter as clock, sleep
from datetime import datetime
from abc import abstractmethod, ABC
from os import getpid, path
from psutil import Process
from _functools import reduce
from time import time
from queue import Queue
from logging import handlers
from sys import version_info

from accasim.utils.reader_class import DefaultReader, Reader
from accasim.utils.misc import CONSTANT, DEFAULT_SIMULATION, DEFAULT_SWF_MAPPER, load_config, clean_results
from accasim.utils.misc import SystemStatus
from accasim.utils.file import path_leaf, save_jsonfile, dir_exists
from accasim.base.event_class import EventManager, AttributeType
from accasim.base.resource_manager_class import Resources, ResourceManager
from accasim.base.scheduler_class import SchedulerBase
from accasim.base.event_class import JobFactory
from accasim.base.additional_data import AdditionalData, AdditionalDataError
from accasim.utils.async_writer import AsyncWriter
import traceback

class SimulatorBase(ABC):
    
    LOG_LEVEL_INFO = 'INFO'
    LOG_LEVEL_DEBUG = 'DEBUG'
    LOG_LEVEL_TRACE = 'TRACE'    
    
    def __init__(self, config_file=None, **kwargs):
        """

        Simulator base constructor

        :param resource_manager: Resource manager class instantiation
        :param reader: Reader class instantiation
        :param job_factory: Job Factory instantiation
        :param _dispatcher: Dispatcher instantiation
        :param config_file: Path to the config file in json format.
        :param \*\*kwargs: Dictionary of key:value parameters to be used in the simulator. It overwrites the current parameters. All parameters will be available on the constant variable

        """        
        self.constants = CONSTANT()
        self.timeout = kwargs.pop('timeout', None)
        self._id = kwargs.pop('id', None)
        self._log_level = kwargs.pop('LOG_LEVEL', self.LOG_LEVEL_INFO)
        self.define_default_constants(config_file, **kwargs)

        self._logger, self._logger_listener = self.define_logger()
        self.real_init_time = datetime.now()
        
        # self.dispatcher = None
        self.reader = None
        self.resource_manager = None
        self.job_factory = None
        self.mapper = None
        self.additional_data = None
               
        if self.constants.OVERWRITE_PREVIOUS:
            self.remove_previous()

    def define_logger(self):
        self._define_trace_logger()
        FORMAT = '%(asctime)-15s %(module)s-%(levelname)s: %(message)s'
        
        queue = Queue(-1)
        queue_handler = handlers.QueueHandler(queue)
        handler = logging.StreamHandler()
        handler.setLevel(self._log_level)
        listener = handlers.QueueListener(queue, handler)
        
        logger_name = 'accasim'
        logger = logging.getLogger(logger_name)    
        logger.addHandler(queue_handler)
        formatter = logging.Formatter(FORMAT)
        handler.setFormatter(formatter)
        logger.setLevel(getattr(logging, self._log_level))
        
        self.constants.load_constant('LOGGER_NAME', logger_name)
        return logger, listener
         
    def _define_trace_logger(self):
        level = logging.TRACE = logging.DEBUG - 5
    
        def log_logger(self, message, *args, **kwargs):
            if self.isEnabledFor(level):
                self._log(level, message, args, **kwargs)
        logging.getLoggerClass().trace = log_logger
    
        def log_root(msg, *args, **kwargs):
            logging.log(level, msg, *args, **kwargs)
        logging.addLevelName(level, "TRACE")
        logging.trace = log_root
        
    @abstractmethod
    def start_simulation(self):
        """

        Simulation initialization

        """
        raise NotImplementedError('Must be implemented!')

    @abstractmethod
    def load_events(self):
        """

        Method that loads the job from a datasource. Check the default implementation in the Simulator class.

        """
        raise NotImplementedError('Must be implemented!')

    def additional_data_init(self, _additional_data):
        """

        Initializes the additional_data classes or set the event manager in the objects

        :param _additional_data: A list of AdditionalData objects or classes

        :return: Return a list with all the AdditionalData objects ready to be executed

        """
        _ad = []
        for ad in _additional_data:
            if not ad:
                raise AdditionalDataError('AdditionalData class/object is None')
                
            if isinstance(ad, AdditionalData):
                ad.set_event_manager(self.mapper)
                _ad.append(ad)
            elif issubclass(ad, AdditionalData):
                _ad.append(ad(self.mapper))
            else:
                raise Exception('Additional data class must be a subclass of the AdditionalData class. Received {}'.format(AdditionalData.__class__))
        return _ad

    def check_request(self, attrs_names):
        """

        Verifies that the job factory attributes can be supported by the system resurces.

        :return: True if attributes are supported, False otherwise.

        """
        _system_resources = self.resource_manager.system_resource_types
        for _res in _system_resources:
            if not (_res in attrs_names):
                print('Resource \'{}\' is not included in the Job dict.'.format(_res))
                return False
        return True

    def generate_enviroment(self, config_path):
        """

        Generated the syntethic system from the config file

        :param config_path: Path the config file

        :return: resource manager object.

        """
        config = load_config(config_path)
        equiv = config.pop('equivalence', {})
        start_time = config.pop('start_time', 0)
        resources = Resources(**config)
        return ResourceManager(resources), equiv, start_time

    def define_filepaths(self, **kwargs):
        """

        Add to the kwargs useful filepaths.

        """
        kwargs['WORKLOAD_FILENAME'] = path_leaf(kwargs['WORKLOAD_FILEPATH'])[1]
        if 'RESULTS_FOLDER_PATH' not in kwargs:
            # New in 3.5 
            filename = stack()[-1].filename
            script_path, script_name = path_leaf(filename)
            rfolder = kwargs.pop('RESULTS_FOLDER_NAME')
            kwargs['RESULTS_FOLDER_PATH'] = path.join(script_path, rfolder)
        dir_exists(kwargs['RESULTS_FOLDER_PATH'], create=True)
        return kwargs

    def set_workload_input(self, workload_path, **kwargs):
        """

        Creates a default reader object

        :param workload_path: Path to the workload
        :param \*\*kwargs: extra arguments

        :return: A reader object

        """
        return DefaultReader(workload_path, **kwargs)

    def prepare_arguments(self, possible_arguments, arguments):
        """

        Verifies arguments for a specific instantation and create the dictionary.

        :Note:

            this method will be moved to misc

        :param possible_arguments: Required arguments.
        :param arguments: Available arguments.

        :return: Dictionary with the corresponding arguments.

        """
        return {k: v for k, v in arguments.items() if k in possible_arguments}

    def define_default_constants(self, config_filepath, **kwargs):
        """

        Defines the default constants of the simulator, and update if the user gives new values.

        :param config_filepath: Path to the config file in json format

        """
        config = DEFAULT_SIMULATION
        for k, v in config.items():
            if k not in kwargs:
                kwargs[k] = v
        if config_filepath:
            for k, v in load_config(config_filepath).items():
                kwargs[k] = v
        kwargs = self.define_filepaths(**kwargs)
        self.constants.load_constants(kwargs)

    def show_config(self):
        """

        Shows the current simulator config

        """
        self._logger.info('Initializing the simulator')
        self._logger.info('Settings: ')
        self._logger.info('\tSystem Configuration file: {}'.format(self.constants.SYS_CONFIG_FILEPATH))
        self._logger.info('\tWorkload file: {}'.format(self.constants.WORKLOAD_FILEPATH))
        self._logger.info('\tResults folder: {}{}.'.format(self.constants.RESULTS_FOLDER_PATH,
                                               ', Overwrite previous files' if self.constants.OVERWRITE_PREVIOUS else ''))
        self._logger.info('\t\t ({}) Dispatching Plan Output. Prefix: {}'.format(self.on_off(self.constants.SCHEDULING_OUTPUT),
                                                                     self.constants.SCHED_PREFIX))
        self._logger.info('\t\t ({}) Statistics Output. Prefix: {}'.format(self.on_off(self.constants.STATISTICS_OUTPUT),
                                                               self.constants.STATISTICS_PREFIX))
        self._logger.info('\t\t ({}) Dispatching Plan. Pretty Print Output. Prefix: {}'.format(
            self.on_off(self.constants.PPRINT_OUTPUT), self.constants.PPRINT_PREFIX))
        self._logger.info('\t\t ({}) Benchmark Output. Prefix: {}'.format(self.on_off(self.constants.BENCHMARK_OUTPUT),
                                                              self.constants.BENCHMARK_PREFIX))
        self._logger.info('Ready to Start')

    def on_off(self, state):
        """

        True: ON, False: OFF
        Just for visualization purposes.

        :param state: State of a constant. True or False

        """
        return 'ON' if state else 'OFF'

    def remove_previous(self):
        """

        To clean the previous results.

        """
        _wouts = [(self.constants.SCHEDULING_OUTPUT, self.constants.SCHED_PREFIX),
                  (self.constants.STATISTICS_OUTPUT, self.constants.STATISTICS_PREFIX),
                  (self.constants.PPRINT_OUTPUT, self.constants.PPRINT_PREFIX),
                  (self.constants.BENCHMARK_OUTPUT, self.constants.BENCHMARK_PREFIX)]

        _paths = [path.join(self.constants.RESULTS_FOLDER_PATH, _prefix + self.constants.WORKLOAD_FILENAME) for
                  state, _prefix in _wouts if state]
        clean_results(*_paths)

    def _clean_simulator_constants(self):
        self.constants.clean_constants()

    def _save_parameters(self, _parameters, filename='simulator_parameters.json'):
        filename_path = path.join(self.constants.RESULTS_FOLDER_PATH, filename)
        _dict = {
            _param: getattr(self.constants, _param) for _param in _parameters
        }
        save_jsonfile(filename_path, _dict)

    def _generated_filepaths(self):
        possible_filepaths = [
            (self.constants.STATISTICS_OUTPUT, self.constants.STATISTICS_PREFIX, path.join(self.constants.RESULTS_FOLDER_PATH, self.constants.STATISTICS_PREFIX + self.constants.WORKLOAD_FILENAME)),
            (self.constants.BENCHMARK_OUTPUT, self.constants.BENCHMARK_PREFIX, path.join(self.constants.RESULTS_FOLDER_PATH, self.constants.BENCHMARK_PREFIX + self.constants.WORKLOAD_FILENAME)),
            (self.constants.SCHEDULING_OUTPUT, self.constants.SCHED_PREFIX, path.join(self.constants.RESULTS_FOLDER_PATH, self.constants.SCHED_PREFIX + self.constants.WORKLOAD_FILENAME)),
            (self.constants.PPRINT_OUTPUT, self.constants.PPRINT_PREFIX, path.join(self.constants.RESULTS_FOLDER_PATH, self.constants.PPRINT_PREFIX + self.constants.WORKLOAD_FILENAME))
        ]
        return {f[1]:f[2] for f in possible_filepaths if f[0]}

class Simulator(SimulatorBase):
    """

    Default implementation of the SimulatorBase class.

    """

    def __init__(self, workload, sys_config, dispatcher, resource_manager=None, reader=None, job_factory=None,
                 additional_data=[], simulator_config=None, overwrite_previous=True,
                 scheduling_output=True, pprint_output=False, benchmark_output=False, statistics_output=True, save_parameters=None,
                 show_statistics=True, **kwargs):
        """

        Constructor of the HPC Simulator class.

        :param workload: Filepath to the workload, it is used by the reader. If a reader is not given, the default one is used.
        :param sys_config: Filepath to the synthetic system configuration. Used by the resource manager to create the system.
        :param dispatcher: Dispatching method
        :param resource_manager: Optional. Instantiation of the resource_manager class.
        :param reader: Optional. Instantiation of the reader class.
        :param job_factory: Optional. Instantiation of the job_factory class.
        :param additional_data: Optional. Array of Objects or Classes of AdditionalData class.
        :param simulator_config: Optional. Filepath to the simulator config. For replacing the misc.DEFAULT_SIMULATION parameters.
        :param overwrite_previous: Default True. Overwrite previous results.
        :param scheduling_output: Default True. Dispatching plan output. Format modificable in DEFAULT_SIMULATION
        :param pprint_output: Default False. Dispatching plan output in pretty print version. Format modificable in DEFAULT_SIMULATION
        :param benchmark_output: Default False. Measurement of the simulator and dispatcher performance.
        :param statistics_output: Default True. Statistic of the simulation.
        :param save_parameters: List of simulation name paremeters to be saved in the target results folder. None or empty for not saving the parameters.
        :param show_statistics: Default True. Show Statistic after finishing the simulation.
        :param \*\*kwargs: Optional parameters to be included in the Constants.

        """ 
        assert(version_info >= (3, 5,)), 'Unsupported python version. Try with 3.5 or newer.' 
        
        kwargs['OVERWRITE_PREVIOUS'] = overwrite_previous
        kwargs['SYS_CONFIG_FILEPATH'] = sys_config
        kwargs['WORKLOAD_FILEPATH'] = workload
        kwargs['SCHEDULING_OUTPUT'] = scheduling_output
        kwargs['PPRINT_OUTPUT'] = pprint_output
        kwargs['BENCHMARK_OUTPUT'] = benchmark_output
        kwargs['STATISTICS_OUTPUT'] = statistics_output
        kwargs['SHOW_STATISTICS'] = show_statistics

        _uargs = []
        
        if not resource_manager:
            resource_manager, equiv, start_time = self.generate_enviroment(sys_config)
            kwargs['equivalence'] = equiv
            kwargs['start_time'] = start_time
        if not job_factory:
            kwargs['job_mapper'] = DEFAULT_SWF_MAPPER
            if not kwargs.get('EXTENDED_JOB_DESCRIPTION', False):
                kwargs['job_attrs'] = self.default_job_description()
            else:
                kwargs['job_attrs'] = self.extended_job_description()
            _jf_arguments = ['job_class', 'job_attrs', 'job_mapper']
            args = self.prepare_arguments(_jf_arguments, kwargs)
            _uargs += _jf_arguments
            job_factory = JobFactory(resource_manager, **args)
        if workload and not reader:
            _reader_arguments = ['max_lines', 'tweak_function', 'equivalence', 'start_time']
            args = self.prepare_arguments(_reader_arguments, kwargs)
            reader = self.set_workload_input(workload, job_factory=job_factory, **args)
            _uargs += _reader_arguments
        
        for _u in _uargs:
            kwargs.pop(_u, None)
        SimulatorBase.__init__(self, config_file=simulator_config, **kwargs)
        
        if not isinstance(additional_data, list):
            assert (isinstance(additional_data, AdditionalData) or issubclass(additional_data,
                                                                               AdditionalData)), 'Only subclasses of AdditionalData class are acepted as additional_data argument '
            additional_data = [additional_data]
        
        assert (isinstance(resource_manager, ResourceManager))
        self.resource_manager = resource_manager
        
        assert (isinstance(dispatcher, SchedulerBase))
        dispatcher.set_resource_manager(resource_manager)
        # self.dispatcher = dispatcher
            
        assert (isinstance(reader, Reader))
        self.reader = reader
                
        assert (isinstance(job_factory, JobFactory))
        self.job_factory = job_factory
        
        additional_data = self.additional_data_init(additional_data)
        self.mapper = EventManager(self.resource_manager, dispatcher, additional_data)

        if save_parameters:
            self._save_parameters(save_parameters)

        if benchmark_output:
            self._usage_writer = AsyncWriter(path=path.join(self.constants.RESULTS_FOLDER_PATH,
                self.constants.BENCHMARK_PREFIX + self.constants.WORKLOAD_FILENAME),
                pre_process_fun=Simulator.usage_metrics_preprocessor)
            self._process_obj = Process(getpid())
        else:
            self._usage_writer = None
            self._process_obj = None

        self.start_simulation_time = None
        self.end_simulation_time = None
        self.max_sample = 2
        self.daemons = {}
        self.loaded_jobs = 0
        self.dispatched_jobs = 0
        self.rejected_jobs = 0

    def monitor_datasource(self, _stop):
        """

        runs continuously and updates the global data
        Useful for daemons

        :param _stop: Signal for stop

        """
        while (not _stop.is_set()):
            self.constants.running_at['current_time'] = self.mapper.current_time
            self.constants.running_at['running_jobs'] = {x: self.mapper.events[x] for x in self.mapper.running}
            sleep(self.constants.running_at['interval'])

    def start_simulation(self, system_status=False, **kwargs):
        """

        Initializes the simulation

        :param init_unix_time: Adjustement for job timings. If the first job corresponds to 0, the init_unix_time must corresponds to the real submit time of the workload. Otherwise, if the job contains the real submit time, init_unix_time is 0.
        :param system_status: Initializes the system status daemon.
        :param system_utilization: Initializes the running jobs visualization using matplotlib.
        :param \*\*kwargs: a 'tweak_function' to deal with the workloads.

        """
        #=======================================================================
        # System status is the main entry point to get access to the current 
        # simulation data. It is used also for the visualization component.
        #=======================================================================
        if system_status:
            functions = {
                'usage_function': self.mapper.usage,
                'availability_function': self.mapper.availability,
                'simulated_status_function': self.mapper.simulated_status,
                'current_time_function': self.mapper.simulated_current_time
            }
            self.daemons['system_status'] = {
                'class': SystemStatus,
                'args': [self.constants.WATCH_PORT, functions],
                'object': None
            }        
        
        # @TODO
        # Add the usage_writer to the daemons array to auto-on/off process
        self._logger_listener.start() 
        if self._usage_writer:
            self._usage_writer.start()
            
        # Starting the daemons
        self.daemon_init()
        
        self.show_config()
        sim_error = None
        try:
            self.start_hpc_simulation(**kwargs)
        except Exception as e:
            sleep(1)
            print('The simulation will be stopped. Reason: {}'.format(e))        
            sim_error = e
        
        [d['object'].stop() for d in self.daemons.values() if d['object']]
                     
        if self._usage_writer:
            self._usage_writer.stop()
            self._usage_writer = None
 
        # @TODO
        self.mapper.stop_writers()
        
        filepaths = self._generated_filepaths()
        self._clean_simulator_constants()
        self._logger_listener.stop()
        if sim_error:
            raise sim_error
        return filepaths

    def start_hpc_simulation(self, **kwargs):
        """

        Initializes the simulation in a new thread. It is called by the start_timulation using its arguments.

        """
        if self.timeout:
            init_sim_time = time()
            ontime = True
        # =======================================================================
        # Load events corresponding at the "current time" and the next one
        # =======================================================================
        event_dict = self.mapper.events
        self.start_simulation_time = clock()
        self.constants.load_constant('start_simulation_time', self.start_simulation_time)

        self._logger.info('Starting the simulation process.')
        self.load_events(self.mapper.current_time, event_dict, self.mapper, self.max_sample)
        events = self.mapper.next_events()

        # =======================================================================
        # Loop until there are not loaded, queued and running jobs
        # =======================================================================
        while events or self.mapper.has_events():
            current_time = self.mapper.current_time
            benchStartTime = clock() * 1000
            self.mapper.release_ended_events(event_dict)
            queued_len = len(events)

            # ===================================================================
            # External behavior
            # ===================================================================
            # self.execute_additional_data()

            schedEndTime = schedStartTime = clock() * 1000

            if events:
                # to_dispatch, rejected = self.dispatcher.schedule(self.mapper.current_time, event_dict, events)
                to_dispatch, rejected = self.mapper.call_dispatcher(event_dict, events)
                for r in rejected:
                    del event_dict[r]
                dispatched_len = len(to_dispatch)
                rejected_len = len(rejected)
                assert(queued_len == dispatched_len + rejected_len), 'Some queued jobs ({}/{}) were not included in the dispatching decision.'\
                    .format(dispatched_len + rejected_len, queued_len)
                self.rejected_jobs += rejected_len
                
                schedEndTime = clock() * 1000
                time_diff = int((schedEndTime - schedStartTime) / 1000)
                
                (n_disp, n_disp_finish, n_post) = self.mapper.dispatch_events(event_dict, to_dispatch, time_diff)
                self.dispatched_jobs += n_disp + n_disp_finish
                
            # ===================================================================
            # Loading next jobs based on Time points
            # ===================================================================
            if len(self.mapper.loaded) < 10:
                sample = self.max_sample if (len(self.mapper.loaded) < self.max_sample) else 2
                self.load_events(current_time, event_dict, self.mapper, sample)
            # ===================================================================
            # Continue with next events
            # ===================================================================
            events = self.mapper.next_events()

            if self.constants.BENCHMARK_OUTPUT:                
                benchEndTime = clock() * 1000
                benchMemUsage = self._process_obj.memory_info().rss / float(2 ** 20)
                scheduleTime = schedEndTime - schedStartTime
                dispatchTime = benchEndTime - benchStartTime - scheduleTime
                
                self._usage_writer.push((current_time, queued_len, benchEndTime - benchStartTime, scheduleTime,
                                        dispatchTime, benchMemUsage))

            if self.timeout and self.timeout <= int(time() - init_sim_time):
                ontime = False
                break
            
        self.end_simulation_time = clock()
        if not self.timeout or self.timeout and ontime:
            assert (self.loaded_jobs == self.dispatched_jobs + self.rejected_jobs), \
                'Loaded {} != dispatched + rejected {} jobs'.format(self.loaded_jobs, self.dispatched_jobs + self.rejected_jobs)

        self.statics_write_out(self.constants.SHOW_STATISTICS, self.constants.STATISTICS_OUTPUT)
        self._logger.info('Simulation process completed.')
        self.mapper.current_time = None
        

    @staticmethod
    def usage_metrics_preprocessor(entry):
        """
        To be used as a pre-processor for AsyncWriter objects applied to usage metrics.
        Pre-processes a tuple of usage metrics containing 6 fields. The fields are the following:

        - time: the timestamp relative to the simulation step
        - queueSize: the size of the queue at the simulation step (before scheduling)
        - stepTime: the total time required to perform the simulation step
        - schedTime: the time related to the scheduling procedure
        - simTime: the remaining time used in the step, related to the simulation process
        - memUsage: memory usage (expressed in MB) at the simulation step

        :param entry: Tuple of data to be written to output
        """
        sep_token = ';'
        bline = sep_token.join([str(v) for v in entry]) + '\n'
        return bline

    def statics_write_out(self, show, save):
        """

        Write the statistic output file

        :param show: True for showing the statistics, False otherwise.
        :param save: True for saving the statistics, False otherwise.

        """
        if not (show or save):
            return
        wtimes = self.mapper.wtimes
        slds = self.mapper.slowdowns
        sim_time_ = 'Simulation time: {0:.2f} secs\n'.format(self.end_simulation_time - self.start_simulation_time)
        disp_method_ = 'Dispathing method: {}\n'.format(self.mapper.dispatcher)
        total_jobs_ = 'Total jobs: {}\n'.format(self.loaded_jobs)
        makespan_ = 'Makespan: {}\n'.format(self.mapper.last_run_time - self.mapper.first_time_dispatch if self.mapper.last_run_time and self.mapper.first_time_dispatch else 'NA')
        if wtimes:
            avg_wtimes_ = 'Avg. waiting times: {:.2f}\n'.format(reduce(lambda x, y: x + y, wtimes) / float(len(wtimes)))
        else:
            avg_wtimes_ = 'Avg. waiting times: NA\n'
            
        if slds:
            avg_slowdown_ = 'Avg. slowdown: {:.2f}\n'.format(reduce(lambda x, y: x + y, slds) / float(len(slds)))
        else:
            avg_slowdown_ = 'Avg. slowdown: NA\n'
            
        if show:
            self._logger.info('\t ' + sim_time_[:-1])
            self._logger.info('\t ' + disp_method_[:-1])
            self._logger.info('\t ' + total_jobs_[:-1])
            self._logger.info('\t ' + makespan_[:-1])
            self._logger.info('\t ' + avg_wtimes_[:-1])
            self._logger.info('\t ' + avg_slowdown_[:-1])
            
        if save:
            _filepath = path.join(self.constants.RESULTS_FOLDER_PATH,
                                   self.constants.STATISTICS_PREFIX + self.constants.WORKLOAD_FILENAME)
            with open(_filepath, 'a') as f:
                f.write(sim_time_)
                f.write(disp_method_)
                f.write(total_jobs_)
                f.write(makespan_)
                f.write(avg_wtimes_)
                f.write(avg_slowdown_)

    def load_events(self, current_time, jobs_dict, mapper, time_samples=2):
        """

        Incremental loading. Load the new jobs into the

        :param current_time: Current simulation time.
        :param jobs_dict: Dictionary of the current load, queued and running jobs
        :param mapper: Job event mapper object
        :param time_samples: Default 2. It load the next two time steps.

        """
        next_tpoints, parsed_jobs = self.reader.next(current_time, time_points=time_samples)
        tmp_dict = {}
        job_list = []

        for next_tpoint in next_tpoints:
            for job in parsed_jobs[next_tpoint]:
                self.loaded_jobs += 1
                tmp_dict[job.id] = job
                job_list.append(job)
        mapper.load_events(job_list)
        jobs_dict.update(tmp_dict)

    def _loaded_jobs(self):
        return sum([len(jobs) for _, jobs in self.mapper.loaded.items()])

    def daemon_init(self):
        """

        Initialization of the simulation daemons. I.e. system_utilization or system_status

        """
        _iter_func = lambda act, next: act.get(next) if isinstance(act, dict) else (
        getattr(act, next)() if callable(getattr(act, next)) else getattr(act, next))
        for _name, d in self.daemons.items():
            _class = d['class']
            if not _class:
                continue
            _args = []
            for _arg in d['args']:
                if isinstance(_arg, tuple):
                    res = reduce(_iter_func, _arg[1].split('.'), self if not _arg[0] else _arg[0])
                    _args.append(res)
                else:
                    _args.append(_arg)
            self.daemons[_name]['object'] = _class(*_args)
            self.daemons[_name]['object'].start()

    def default_job_description(self):
        """

        Method that returns the minimal attributes of a job. Default values: ID, Expected Duration, CORE and MEM.
        
        :return: Array of Attributes

        """
        # Attribute to identify the user
        user_id = AttributeType('user_id', int)

        # New attributes required by the Dispatching methods.
        expected_duration = AttributeType('expected_duration', int)
        queue = AttributeType('queue', int)

        # Default system resources: core and mem.
        total_cores = AttributeType('core', int)
        total_mem = AttributeType('mem', int)

        return [total_cores, total_mem, expected_duration, queue, user_id]


    def extended_job_description(self):
        """

        Method that returns extra attributes of a job. By default it includes the default attributes: ID, Expected Duration, CORE and MEM.
        
        :return: Array of Attributes

        """
        description = self.default_job_description()
        
        executable = AttributeType('executable', str, optional=True)
        group_id = AttributeType('group_id', int, optional=True)
        status = AttributeType('status', int, optional=True)
        partition = AttributeType('partition', int, optional=True)
        
        return description + [executable, group_id, status]



"""
MIT License

Copyright (c) 2017 cgalleguillosm

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import logging
from copy import deepcopy
from accasim.utils.misc import CONSTANT, FrozenDict
from datetime import datetime

class Resources:
    """
    
        resources class: Stablish the resources, allocate and release their use.
        
    """
    ON = 1
    OFF = 0
    
    def __init__(self, groups, resources, **kwargs):
        """
        
        :param groups: define the groups of resources. i.e: {'group_0': {'core': 4, 'mem': 10}, .. }
        :param resources: Stablish the available resources of the system, in terms of number of previous groups. i.e: {'group_0': 32}, This will set 32 nodes of the group_0
        :param \*\*kwargs:
            - node_prefix: This will set the prefix of the node name. The default name is 'node', this name is followed by _(i) where i corresponds to the ith loaded node.
            - available_prefix: This will set the prefix of the available resources. Internal use
            - used_prefix: This will set the prefix of the used resources. Internal use
        
        """
        self.SYSTEM_CAPACITY_TOTAL = None
        self.SYSTEM_CAPACITY_NODES = None
        self.GROUPS = None
        self.NODE_LIST = []
        
        self._definition = tuple([{'nodes': q, 'resources':groups[k]} for k, q in resources.items()])
        self._resources = {}
        self._current_capacity = {}
        self._resources_status = {}
        self._system_resource_types = []
        
        self._node_prefix = kwargs.pop('node_prefix', 'node_') 
        
        # List all attributes of all groups
        for group_name, group_values in groups.items():
            self._system_resource_types += filter(lambda x: x not in self._system_resource_types, list(group_values.keys()))
        
        # Here the current usage is maintained
        self._total_resources = {r: 0 for r in self._system_resource_types}
        
        #=======================================================================
        # Create the corresponding group attributes and add 0 to absent attributes.
        # This is performed in case that the user doesn't assign an absent attribute in the system config.
        # For instance when a group has gpu and an another group hasn't and that attribute must be 0. 
        #=======================================================================
        _groups = {}
        for group_name, group_values in groups.items():
            resource_group = { attr: group_values.get(attr, 0) for attr in self._system_resource_types }
            self._define_group(_groups, group_name, resource_group)
        self.GROUPS = FrozenDict(**_groups)

        #=======================================================================
        # Define system capacity and usage of the system
        #=======================================================================
        _node_capacity = {res:0 for res in self._system_resource_types}
        _nodes_capacity = {}
        _system_capacity = deepcopy(_node_capacity)

        j = 0        
        for group_name, q in resources.items():
            for res, value in self.GROUPS[group_name].items():
                _system_capacity[res] += value * q
                
            for i in range(q):
                _node_name = '{}{}'.format(self._node_prefix, j + 1)
                _attrs_values = self.GROUPS[group_name]
                
                _nodes_capacity[_node_name] = deepcopy(_attrs_values)
                self._current_capacity[_node_name] = deepcopy(_attrs_values)
                self._resources[_node_name] = deepcopy(_node_capacity)
                self._resources_status[_node_name] = self.ON
                self.NODE_LIST.append(_node_name)   
                j += 1

        self.SYSTEM_CAPACITY_NODES = FrozenDict(**_nodes_capacity)
        self.SYSTEM_CAPACITY_TOTAL = FrozenDict(**_system_capacity)

    def _define_group(self, _groups, name, group):
        """
        
         Internal method for defining groups of resources.
         
         :param name: Name of the group
         :param group: Values of the group. As defined in the system config.  
        
        """
        assert(isinstance(group, dict))
        assert(name not in _groups), 'Repreated name group: {}. Select another one.'.format(name)
        _groups[name] = group

    def allocate(self, node_name, **kwargs):
        """
        
        Method for job allocation. It receives the node name and the resources to be used.
        
        :param node_name: Name of the node to be updated.
        :param \*\*kwargs: Dictionary of the system resources and its values to be used. 
        
        """
        assert(self._resources), 'The resources must be setted before jobs allocation'
        assert(self._resources_status[node_name] == self.ON), 'The Node {} is {}, it is impossible to allocate any job'
        _capacity = self.SYSTEM_CAPACITY_NODES[node_name]
        _used_resources = self._resources[node_name]
        _done = []
        _full_usage = {}
        for res, v in kwargs.items():
            _rem_res = _capacity[res] - _used_resources[res]  # _used_resources[avl_key] - _used_resources[used_key]
            try:
                assert(v <= _rem_res), 'The event requested {} {}, but there are only {} available.'.format(v, res, _rem_res)
                _used_resources[res] += v
                # Update totals
                self._total_resources[res] += v
                self._current_capacity[node_name][res] -= v
                _done.append((res, v))
            except AssertionError as e:
                while _done:
                    key, req = _done.pop()
                    _used_resources[key] -= req
                    # Update totals
                    self._total_resources[key] -= req
                    self._current_capacity[node_name][key] += req
                return False, e
        return True, 'OK'

    def release(self, node_name, **kwargs):
        """
        
        Method for allocation release. It receives the node name and the resources to be released.
        
        :param node_name: Name of the node to be updated.
        :param \*\*kwargs: Dictionary of the system resources and its values to be released. 
        
        """
        assert(self._resources), 'The resources must be setted before release resources'
        assert(self._resources_status[node_name] == self.ON), 'The Node {} is {}.'
        _resources = self._resources[node_name]
        for _res, v in kwargs.items():
            _resources[_res] -= v
            # Update totals
            self._total_resources[_res] -= v
            self._current_capacity[node_name][_res] += v
            assert(_resources[_res] >= 0), 'The event was request to release {} {}, but there is only {} available. It is impossible less than 0 resources'.format(v, _res, _resources['%s%s' % (self.used_prefix, _res)])
        
    def availability(self):
        """
        
        Returns the current system availablity. It just return nodes that are ON.
        
        :return: Return a dictionary with the system availability. In terms of {node: {resource: value}}
        
        """
        return {node: {k:v for k, v in self._current_capacity[node].items()} for node in self.NODE_LIST if self._resources_status[node] == self.ON}

    def usage(self, type=None):
        """
        
        System usage calculation
        @todo: Use NODE_LIST instead items
        
        :return: Return a string of the system usage 
        
        """
        _str = "System usage: "
        _str_usage = []
        
        if not type:
            for res, value in self.SYSTEM_CAPACITY_TOTAL.items():
                if value > 0:
                    _str_usage.append("{}: {:.2%}".format(res, self._total_resources[res] / value))
                else:
                    _str_usage.append("{}: -".format(res))
            return (_str + ', '.join(_str_usage))
        elif type == 'dict':
            return {
                res: (self._total_resources[res] / value) * 100                
                for res, value in self.SYSTEM_CAPACITY_TOTAL.items()
            }
        else:
            raise NotImplementedError()

    def system_capacity(self, type='total'):
        """
        
        :param type: 
            'total' to return the total per resource type
            'nodes' to return the capacity of nodes
                        
        
        :return: Return system capacity 
        
        """
        if type == 'total':
            return self.SYSTEM_CAPACITY_TOTAL
        elif type == 'nodes':
            return self.SYSTEM_CAPACITY_NODES
        raise ResourceError('System Capacity: \'{}\' type not defined'.format(type))
    
    def resource_manager(self):
        """
        
        Instantiation of the resource manager object
        
        :return: Resource manager object. 
        
        """
        return ResourceManager(self)

    def __str__(self):
        """
        @todo: Use NODE_LIST instead items
        """
        _str = "Resources:\n"
        for node in self.NODE_LIST:
            attrs = self._resources[node]
            formatted_attrs = ""
            for attr in self._system_resource_types:
               formatted_attrs += '{}: {}/{}, '.format(attr, attrs[attr], self.SYSTEM_CAPACITY_NODES[node][attr])
            _str += '- {}: {}\n'.format(node, formatted_attrs)
        return _str
    
    def system_groups(self):
        """
            Returns the available system groups.
        """
        return self.GROUPS

class ResourceManager:

    def __init__(self, _resource):
        """
        
        Constructor for Resource Manager.
        This class handles the resources through Allocation and Release methods.
        
        :param _resource: An instance of the resources class. It defines the system capacity.  
        
        """
        assert(isinstance(_resource, Resources)), ('Only {} class is acepted for resources'.format(Resources.__name__))
        self._resources = _resource
        self._running_jobs = {}
        self._logger = logging.getLogger('accasim')

    def allocate_event(self, event, node_names):
        """
        
        Method for job allocation. It uses the event request to determine the resources to be allocated.
        
        :param event: Job event object.
        :param node_names: List of nodes where the job will be allocated.  
        
        :return: Tuple: First element True if the event was allocated, False otherwise. Second element a message. 
        """
        self._logger.trace('Allocating {} in nodes {}'.format(event.id, ', '.join([node for node in node_names])))
        _requested_res = event.requested_resources
        _attrs = _requested_res.keys()
        
        _allocation = {}
        for node in node_names:
            if not(node in _allocation):
                _allocation[node] = {_attr: _requested_res[_attr] for _attr in _attrs}
                continue
            for _attr in _attrs:
                _allocation[node][_attr] += _requested_res[_attr]
        
        _allocated = True
        _rollback = []

        for node_name, values in _allocation.items():
            done, message = self._resources.allocate(node_name, **values)
            if done:
                _rollback.append((node_name, values))
            else:
                self._logger.error('Rollback for {}: {}'.format(event.id, _rollback + [(node_name, values)]))
                _allocated = False
                break
        
        if _allocated:
            self._running_jobs[event.id] = _allocation
        else:
            while _rollback:
                node_name, values = _rollback.pop()
                self._resources.release(node_name, **values)
               
        return _allocated, message

    def remove_event(self, id):
        """
        
        Method for job release. It release the allocated resources on the specific nodes.
        
        :param id: Job Id 
        
        """
        for node_name, values in self._running_jobs.pop(id).items():
            self._resources.release(node_name, **values)

    def node_resources(self, *args):
        """
        
        :param \*args: list of node names 
        
        Print nodes and its resources 
        
        """
        for arg in args:
            print(arg, self._resources._resources[arg])

    
    @property
    def current_availability(self):
        """
        
        :return: Return system availability
        
        """        
        return self._resources.availability()

    @property
    def resource_types(self):
        """
        
        :return: Return resource types of the system
        
        """
        return self._resources._system_resource_types

    @property
    def node_names(self):
        """
        
        :return: Return node names
        
        """
        return self._resources.NODE_LIST
    
    def total_resources(self, *args):
        """
        
        Return the total system resource for the required argument. The resource has to exist in the system. 
        If no arguments is proportioned all resources are returned.
        @todo: Use NODE_LIST instead items
        
        :param \*args: Depends on the system configuration. But at least it must have ('core', 'mem') resources.
            
        :return: Dictionary of the resources and its values.          
        
        """
        _resources = self.system_capacity('total')
        if not args or len(args) == 0:
            return {k: v for k, v in _resources.items()}
        avl_types = {}
        for arg in args:
            assert(arg in _resources), '{} is not a resource of the system. Available resource are {}'.format(arg, self.resource_types) 
            avl_types[arg] = _resources[arg]
        return avl_types

    def groups_available_resource(self, _key=None):
        """
        
        :param _key: None for values of all types for all groups. Giving a specific key will return the resource for the specific type
        :todo: Use NODE_LIST instead items
        
        :return: Dictionary of {group{type: value}}   
        
        """
        _resources = self._resources.system_groups()
        if not _key:
            return _resources
        _resource_key = self._resources.available_resource_key(_key)
        return {_group:_v[_resource_key]  for _group, _v in self._resources.items()} 
    
    def system_resource_types(self):
        return self._resources._system_resource_types
    
    def system_capacity(self, type):
        """        
        :param type: 
            'total' to return the total per resource type
            'nodes' to return the capacity of nodes            
        
        :return: Return system capacity 
        """
        return self._resources.system_capacity(type)
    
    def system_resources(self):
        return self._resources 
    
    @property    
    def current_usage(self):
        return self._resources.usage()
    
    @property
    def current_allocations(self):
        return self._running_jobs

class ResourceError(Exception):
    pass  



"""
MIT License

Copyright (c) 2017 cgalleguillosm, AlessioNetti

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import logging

from random import seed
from sys import maxsize 
from sortedcontainers import SortedList
from abc import abstractmethod, ABC
from accasim.base.resource_manager_class import ResourceManager


class AllocatorBase(ABC):
    """
    
    The base abstract interface all allocators must comply to.
    
    """

    MAXSIZE = maxsize

    def __init__(self, _seed, **kwargs):
        """
    
        Allocator constructor (based on scheduler)

        :param seed: Seed if there is any random event
        :param res_man: resource manager for the system.
        :param kwargs: Nothing for the moment
                 
        """
        seed(_seed)
        self.avl_resources = None
        self.node_names = None
        self.resource_manager = None
        
        self._logger = logging.getLogger('accasim')
        # The list of resource types that are necessary for job execution. Used to determine wether a node can be used
        # for allocation or not
        self.nec_res_types = ['core', 'mem']


    @abstractmethod
    def get_id(self):
        """
    
        Abstract method. Must be implemented by the subclass. 
        Must return the identification of the allocator. 
        
        :return: Allocator identification (for instance its name).    
    
        """
        raise NotImplementedError

    @abstractmethod
    def set_resources(self, res):
        """
    
        Abstract method. Must be implemented by the subclass.
        This method sets the internal reference to the dictionary of available resources in the system.
        If the reference points to a list used also outside of this class, the object should be deepcopied.
        
        If necessary, the resources are also sorted.
            
        :param res: the list of currently available resources in the system.       
    
        """
        raise NotImplementedError
    
    def get_resources(self):
        """
            Returns the internal reference to the dictionary of available resources in the system. 
            It includes the last virtual allocations.
        """
        return self.avl_resources

    @abstractmethod
    def set_attr(self, **kwargs):
        """
    
        Abstract method. Must be implemented by the subclass.
        Method used to set internal parameters and meta-data for the allocator.
        
        Its behavior depends on the specific allocator that is being used, and some arguments may be discarded.
        
        :param kwargs: the internal parameters to be set, depending on the allocator
    
        """
        raise NotImplementedError

    @abstractmethod
    def allocating_method(self, es, cur_time, skip=False, reserved_time=None, reserved_nodes=None):
        """
    
        Abstract method. Must be implemented by the subclass.
        This method must try to allocate the scheduled events contained in es. It will stop as soon as an event cannot
        be allocated, to avoid violations of the scheduler's priority rules, or proceed with other events depending
        on the skip parameter.
        
        The method must support both list of events for es, in which case it will return a list, or single events.
        If there is at least one successful allocation, avl_resources is updated and sorted again efficiently.

        :param es: the event(s) to be allocated
        :param cur_time: current time, needed to build the schedule list
        :param skip: determines if the allocator can skip jobs
        :param reserved_time: beginning of the next reservation slot (used for backfilling)
        :param reserved_nodes: nodes already reserved (used for backfilling)

        :return: a list of assigned nodes of length e.requested_nodes, for all events that could be allocated. The list is in the format (time,event,nodes) where time can be either cur_time or None.
        
        """
        raise NotImplementedError
    
    def allocate(self, es, cur_time, skip=False, reserved_time=None, reserved_nodes=None):
        """
    
        This is the method that is called by the Scheduler to allocate the scheduled jobs. First, It verifies the data consistency and availability, 
        and then call to the implemented allocation policy.   
        
        
        :param es: the event(s) to be allocated
        :param cur_time: current time, needed to build the schedule list
        :param skip: determines if the allocator can skip jobs
        :param reserved_time: beginning of the next reservation slot (used for backfilling)
        :param reserved_nodes: nodes already reserved (used for backfilling)
        :param debug: Debugging flag
        
        :return: the return of the implemented allocation policy.

        """
        assert(self.resource_manager is not None), 'The resource manager is not defined. It must defined prior to run the simulation.'
        self._logger.debug('{}: {} queued jobs to be considered in the dispatching plan'.format(cur_time, len(es) if isinstance(es, (list, tuple, SortedList)) else 1))

        # Update current available resources
        self.set_resources(self.resource_manager.current_availability)
        dispatching_decision = self.allocating_method(es, cur_time, skip=skip, reserved_time=reserved_time, reserved_nodes=reserved_nodes)
        
        return dispatching_decision 
    
    def set_resource_manager(self, _resource_manager):
        """
        Internally set the resource manager to deal with resource availability.
        
        :param _resource_manager: A resource manager instance or None. If a resource manager is already instantiated,
             it's used for set internally set it and obtain the system capacity for dealing with the request verifications.
             The dispathing process can't start without a resource manager. 
             
        """
        assert isinstance(_resource_manager, ResourceManager), 'Resource Manager not valid for scheduler'
        self.resource_manager = _resource_manager
        self._define_mappers()
            
    def _define_mappers(self):
        if not self.node_names:
            self.node_names = self.resource_manager.node_names

    def __str__(self):
        """
        
            Retrieves the identification of the allocator.
        
        """
        return self.get_id()

class FirstFit(AllocatorBase):
    """
    
    A simple First-Fit allocator. Does not sort the resources.
        
    This allocator supports both single events and lists of events. It also
     supports backfilling. No sorting of the resources is done, so they are
     considered as they are given in input.
    
    """

    name = 'FF'

    def __init__(self, seed=0, **kwargs):
        """
    
        Constructor for the class.
        
        :param seed: seed for random events (not used)
        :param resource_manager: reference to the system resource manager
        :param kwargs: None at the moment
    
        """
        AllocatorBase.__init__(self, seed, **kwargs)    
        self.sorted_keys = None    

    def get_id(self):
        return self.__class__.__name__

    def set_resources(self, res):
        """
    
        Sets in the internal variable avl_resources the current available resources for the system. It also sorts
        them, if the sort_resources method is implemented.
        
        :param res: the list of currently available resources for the system
    
        """
        self.avl_resources = res
        self._adjust_resources()

    def set_attr(self, **kwargs):
        """
    
        Method used to set internal parameters and meta-data for the allocator.

        Its behavior depends on the specific allocator that is being used, and some arguments may be discarded.
        It is not actively used in this simple allocator (for the moment).

        :param kwargs: None for the moment
    
        """
        pass

    def allocating_method(self, es, cur_time, skip=False, reserved_time=None, reserved_nodes=None):
        """
    
        Given a job list es, this method searches for a suitable allocation for as many jobs as possible.
        
        In normal allocation, the method stops as soon as an event in the list cannot be allocated. In this case,
        ths list of already allocated jobs is returned. This is done to be coherent with the scheduler's rules.
        As an alternative, the skip parameter can be supplied to allow the scheduler to skip unallocated jobs.
        This method also support backfilling schedule. In this case, the backfilling parameters are supplied,
        and the allocator tries to fit jobs without delaying the reserved job. In this second case,
        the method does not stop when a job cannot be allocated, but simply skips it.
        
        es can be a list or a single event object. The return type (list or single tuple) changes accordingly.
        
        :param es: the event(s) to be allocated
        :param cur_time: current time, needed to build the schedule list
        :param skip: determines if the allocator can skip jobs
        :param reserved_time: beginning of the next reservation slot (used for backfilling)
        :param reserved_nodes: nodes already reserved (used for backfilling)

        :return: a list of assigned nodes of length e.requested_nodes, for all events that could be allocated. The list is in the format (time,event,nodes) where time can be either cur_time or None.
    
        """
        if not isinstance(es, (list, tuple, SortedList)):
            listAsInput = False
            es = [es]
        else:
            listAsInput = True
        
        allocation = []
        success_counter = 0

        for e in es:
            requested_nodes = e.requested_nodes
            requested_resources = e.requested_resources

            # If the input arguments relative to backfilling are not supplied, the method operates in regular mode.
            # Otherwise, backfilling mode is enabled, allowing the allocator to skip jobs and consider the reservation.
            nodes_to_discard = self._compute_reservation_overlaps(e, cur_time, reserved_time, reserved_nodes)
            backfilling_overlap = False if len(nodes_to_discard) == 0 else True

            assigned_nodes = []
            nodes_left = requested_nodes
                
            for node in self.sorted_keys:
                # Check whether the given node belongs to the list of reserved nodes, in backfilling.
                # If it does, the node is discarded.
                if backfilling_overlap:  # and node in nodes_to_discard:
                    continue
                # First, compute the number of job units fitting in the current node, and then update the job assignment
                resources = self.avl_resources[node]
                fits = self._event_fits_node(resources, requested_resources)                   
                
                if fits == 0:
                    continue
                
                if nodes_left <= fits:
                    assigned_nodes += [node] * nodes_left
                    nodes_left = 0
                else:
                    assigned_nodes += [node] * fits
                    nodes_left -= fits
                
                # Break the loop when the request is satisfied
                if nodes_left == 0:
                    break

            # If, after analyzing all nodes, the allocation is still not complete, the partial allocation
            # is discarded.            
            if nodes_left > 0:
                assigned_nodes = []
            assert len(assigned_nodes) in (0, requested_nodes,), 'Requested' + str(requested_nodes) + ' got ' + str(len(assigned_nodes))

            # If a correct allocation was found, we update the resources of the system, sort them again, and
            # add the allocation to the output list.
            if assigned_nodes:
                allocation.append((cur_time, e.id, assigned_nodes))
                # print(e.id, e.requested_nodes, e.requested_resources, len(assigned_nodes), assigned_nodes)
                self._update_resources(assigned_nodes, requested_resources)
                # Sort keys if it is necessary (BF)
                self._adjust_resources(assigned_nodes)
                success_counter += 1
                self._logger.trace('Allocation successful for event {}'.format(e.id))
            #===================================================================
            # If no correct allocation could be found, two scenarios are possible: 
            #     1) normally, the allocator stops here and returns the jobs allocated so far 
            #     2) if the skip parameter is enabled, the job is just skipped, and we proceed with the remaining ones.
            #===================================================================
            else:
                self._logger.trace('Allocation failed for event {} with {} nodes left'.format(e.id, nodes_left))
                allocation.append((None, e.id, []))
                if not skip:
                    # if jobs cannot be skipped, at the first allocation fail all subsequent jobs fail too
                    for e in es[(success_counter + 1):]:
                        allocation.append((None, e.id, []))
                    self._logger.trace('Cannot skip jobs, {} additional pending allocations failed {}'.format(len(es) - success_counter - 1, es[success_counter:]))
                    self._logger.trace('')
                    break
        self._logger.trace('{}/{} successful allocations of events'.format(success_counter, len(es)))
        return allocation if listAsInput else allocation[0]

    def _compute_reservation_overlaps(self, e, cur_time, reserved_time, reserved_nodes):
        """
    
        This method considers an event e, the current time, and a list of reservation start times with relative
        reserved nodes, and returns the list of reserved nodes that cannot be accessed by event e because of overlap.
        
        :param e: the event to be allocated
        :param cur_time: the current time
        :param reserved_time: the list (or single element) of reservation times
        :param reserved_nodes: the list of lists (or single list) of reserved nodes for each reservation
        
        :return: the list of nodes that cannot be used by event e
    
        """
        if reserved_time is None or reserved_nodes is None:
            return []
        else:
            if not isinstance(reserved_time, (list, tuple)):
                if cur_time + e.expected_duration > reserved_time:
                    return reserved_nodes
                else:
                    return []
            else:
                overlap_list = []
                for ind, evtime in enumerate(reserved_time):
                    if cur_time + e.expected_duration > evtime:
                        overlap_list += reserved_nodes[ind]
                return list(set(overlap_list))

    def _update_resources(self, reserved_nodes, requested_resources):
        """
    
        Updates the internal avl_resources list after a successful allocation.
        
        :param reserved_nodes: the list of nodes assigned to the allocated job
        :param requested_resources: the list of resources requested by the job per each node
    
        """
        for node in reserved_nodes:
            resource = self.avl_resources[node]
            for attr, v in requested_resources.items():
                if v == 0:
                    continue
                cur_q = resource[attr]
                assert cur_q - v >= 0, 'In node {}, the resource {} is going below to 0'.format(node, attr)
                resource[attr] -= v

    def _adjust_resources(self, nodes=None):
        """

        Method which must sort the node list at the beginning and after a successful allocation. 
        It must sort the self.sorted_keys attribute.

        """
        if not nodes:
            self.sorted_keys = []
            for node in self.node_names:
                add = True
                for res, avl in self.avl_resources[node].items():
                    if res in self.nec_res_types and avl == 0:
                        add = False
                        break
                if add:
                    self.sorted_keys.append(node)
        else:
            to_remove = set()
            for node in nodes:
               for res, avl in self.avl_resources[node].items():
                   if res in self.nec_res_types and avl == 0:
                       to_remove.add(node)
                       break
            for node in to_remove:
                self.sorted_keys.remove(node)
                

    def _event_fits_node(self, resources, requested_resources):
        _fits = self.MAXSIZE

        for res_type, req in requested_resources.items():
            # The current res_type isnt requested 
            if req == 0:
                continue
            fit = resources[res_type] // req
            # The current res_type cannot be fitted in this node, returning without checking the remaining resource types.
            if fit == 0:
                return 0
            # We maintain the minimum value 
            if _fits > fit:
                _fits = fit
        return _fits

class BestFit(FirstFit):
    """
    
    Best-Fit Allocator
    It is an allocator which sorts the nodes basing on the amount of free resources, trying to consolidate.
        
    The less the available resources, the higher the priority.
    The allocator is based on ffp_alloc, changing only the sort and adjust methods.   
    
    """
    
    name = 'BF'

    def __init__(self, seed=0, **kwargs):
        """
        
        Constructor for the class.

        :param seed: seed for random events (not used)
        :param resource_manager: reference to the system resource manager
        :param kwargs: None at the moment
        
        """
        FirstFit.__init__(self, seed)

        self.ranking = lambda x: sum(self.avl_resources[x].values())
        """
        
            Defines the ranking operator for sorting. Must use the self.avl_resources argument
            (the available resource dictionary). x represents a key.
        
        """
        
    def _adjust_resources(self, nodes=None):
        """

        Method which must sort the node list at the beginning and after a successful allocation. 
        It must sort the self.sorted_keys attribute.

        """
        if not nodes:
            self.sorted_keys = []
            for node in self.node_names:
                add = True
                for res, avl in self.avl_resources[node].items():
                    if res in self.nec_res_types and avl == 0:
                        add = False
                        break
                if add:
                    self.sorted_keys.append(node)
        else:
            to_remove = set()
            for node in nodes:
               for res, avl in self.avl_resources[node].items():
                   if res in self.nec_res_types and avl == 0:
                       to_remove.add(node)
                       break
            for node in to_remove:
                self.sorted_keys.remove(node)
        self.sorted_keys.sort(key=self.ranking, reverse=False)



"""
MIT License

Copyright (c) 2017 cgalleguillosm

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from abc import ABC, abstractclassmethod

from accasim.utils.file import file_exists 

class WorkloadWriter(ABC):
    
    def __init__(self, path, overwrite=False, append=False):
        """
        Abstract class used to write workload files.

        :param path: Path to the target file
        :param overwrite: If True, any existing files with the same name will be overwritten
        :param append: If True, the new workload will be appended to a file with the same name, if it exists
        """
        exists = file_exists(path, True)
        if exists and not (overwrite or append):
            raise Exception('File already exists. Overwrite option is False. Set True to overwrite or change the filename/filepath.')
        if overwrite and append:
            raise Exception('Only one mode (append or overwrite) can be True. ')
        
        mode = None     
        if overwrite:
            mode = 'w'
        elif append:
            mode = 'a'
        
        assert(mode), 'File exists and the overwrite and append modes are False'
        self.file = open(path, mode)
        
    def add_newline(self, job_dict):
        """
        Writes a new line corresponding to a job dictionary given as input

        :param job_dict: The input job dictionary
        """
        line = self.process_dictionary(job_dict)
        if line[-1] != '\n':
            line += '\n'
        self.file.write(line)

    @abstractclassmethod
    def process_dictionary(self, job_dict):
        """
        This method must convert the job dictionary to a string formatted in a specific way, according to the implementation

        :param job_dict: Dictionary related to one specific job
        :return: The string corresponding to the input dictionary
        """
        raise NotImplementedError()
    
    def close_file(self):
        """
        Closes the output file stream
        """
        self.file.close()
    
    def __del__(self):
        """
        If present, closes the file stream.
        """
        if hasattr(self, 'file'):
            self.close_file()


class DefaultWriter(WorkloadWriter):
    """
    Implementation of the WorkloadWriter class targeted at writing workload files in SWF format
    """

    # A list of attributes used to identify and form SWF job entries
    JOB_NUMBER = ('job_number', 0)
    SUBMIT_TIME = ('submit_time', 0)
    WAIT_TIME = ('wait_time', -1)
    DURATION = ('duration', 0)
    ALLOCATED_PROCESSORS = ('allocated_processors', 0)
    AVG_CPU_TIME = ('avg_cpu_time', -1)
    USED_MEMORY = ('used_memory', 0)
    REQUESTED_NUMBER_PROCESSORS = ('requested_number_processors', -1)
    REQUESTED_TIME = ('requested_time', -1)
    REQUESTED_MEMORY = ('requested_memory', -1)
    STATUS = ('status', 1)
    USER_ID = ('user_id', -1)
    GROUP_ID = ('group_id', -1)
    EXECUTABLE_NUMBER = ('executable_number', -1)
    QUEUE_NUMBER = ('queue_number', -1)
    PARTITION_NUMBER = ('partition_number', -1)
    PRECEDING_JOB_NUMBER = ('preceding_job_number', -1)
    THINK_TIME_PREJOB = ('think_time_prejob', -1)
    SWF_ATTRIBUTES = [
        JOB_NUMBER, SUBMIT_TIME, WAIT_TIME, DURATION, ALLOCATED_PROCESSORS,
        AVG_CPU_TIME, USED_MEMORY, REQUESTED_NUMBER_PROCESSORS, REQUESTED_TIME,
        REQUESTED_MEMORY, STATUS, USER_ID, GROUP_ID, EXECUTABLE_NUMBER,
        QUEUE_NUMBER, PARTITION_NUMBER, PRECEDING_JOB_NUMBER, THINK_TIME_PREJOB
    ]
    
    def __init__(self, path, max_time=14400, overwrite=False, append=False):
        """
        Constructor for the class

        :param path: Path of the target workload file
        :param max_time: Represents a wall-time value for jobs to be used in the workload
        :param overwrite: If True, any existing files with the same name will be overwritten
        :param append: If True, the new workload will be appended to a file with the same name, if it exists
        """
        WorkloadWriter.__init__(self, path, overwrite, append)
        self.max_time = max_time
        
    def process_dictionary(self, job_dict):
        """
        Converts a job dictionary to a string to be written in the workload.

        :param job_dict: The job dictionary
        :return: A properly formatted SWF string entry
        """
        line = [str(self.prepare_data(job_dict, attr_name, default_value)) for attr_name, default_value in self.SWF_ATTRIBUTES]
        return '\t'.join(line) 
        
    def prepare_data(self, job_dict, attr_name, default_value):
        """
        Method used to prepare specific SWF attributes, by converting entries from the original job dictionary

        :param job_dict: The job dictionary
        :param attr_name: The name of the attribute to be processed
        :param default_value: Default value to be used for the attribute, if not present in job_dict
        :return: The processed value for attr_name, if present, or default_value otherwise
        """
        total_processors = job_dict['resources']['core'] * job_dict['nodes']
        total_mem = job_dict['resources']['mem'] * job_dict['nodes']
        
        job_dict['requested_number_processors'] = total_processors 
        job_dict['allocated_processors'] = total_processors
        job_dict['requested_memory'] = total_mem
        job_dict['used_memory'] = total_mem
        if not job_dict['requested_time']:
            job_dict['requested_time'] = self.max_time
        
        if attr_name in job_dict:
            return job_dict[attr_name]
        return default_value


"""
MIT License

Copyright (c) 2017 cgalleguillosm

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import warnings
import json
import time

from abc import ABC
from scipy import stats as _statistical_distributions
from scipy.stats import percentileofscore
from statistics import pstdev
from numpy import searchsorted, ndarray
from math import exp, log
from random import random
from sortedcontainers import SortedList
from _functools import reduce

from accasim.base.resource_manager_class import Resources
from accasim.utils.misc import obj_assertion, str_datetime, load_config
from accasim.utils.data_fitting import DistributionFitting
from accasim.experimentation.schedule_writer import WorkloadWriter, DefaultWriter
from accasim.experimentation.schedule_parser import WorkloadFileReader
from accasim.utils.reader_class import DefaultTweaker



class Generator(ABC):
    
    def __init__(self, distributions):
        """
        
        :param distributions:
        """
        if not distributions:
            distributions = [dist for dist in _statistical_distributions.__all__ if isinstance(getattr(_statistical_distributions, dist), _statistical_distributions.rv_continuous)]
        self.distribution_fit = DistributionFitting(distributions)

    def dist_cdf(self, x, dist_name, dist_param, optional):
        """

        :param x:
        :param dist_name:
        :param dist_param:
        :param optional:
        :return:
        """
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore')
            dist = getattr(_statistical_distributions, dist_name)
            return dist.cdf(x, *dist_param, **optional)

    def dist_rand(self, dist_name, dist_param, optional, size=1):
        """

        :param dist_name:
        :param dist_param:
        :param optional:
        :return:
        """
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore')
            dist = getattr(_statistical_distributions, dist_name)
            return dist.rvs(*dist_param, size=size, **optional)
        
    def hist_rand(self, dist_name, dist_param, optional):
        """

        :param dist_name:
        :param dist_param:
        :param optional:
        :return:
        """
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore')
            dist = getattr(_statistical_distributions, dist_name)
            return dist.ppf(random(), *dist_param, **optional)


    def _generate_dist_params(self, data, save=False, filepath=None):
        """

        :param data:
        :param save:
        :param filepath:
        :return:
        """
        ((dist_name, sse, param), bins, hist) = self.distribution_fit.auto_best_fit(data)
        if sse < 0:
            raise Exception('SSE negative.')
        params = {
            'dist_name': dist_name,
            'dist_param': param[:-2],
            'optional': {
                'loc': param[-2],
                'scale': param[-1]
            }
        }
        return (params, bins, hist,) 
        
    def _save_parameters(self, filepath, dist_params, **kwargs):
        """

        :param filepath:
        :param dist_params:
        :param kwargs:
        :return:
        """
        if not filepath:
            filepath = 'dist-params_{}.json'.format(int(time.time()))
        
        gen_parameters = {'params': dist_params}
        for k, v in kwargs.items():
            if isinstance(v, ndarray):
                v = v.tolist()
            gen_parameters[k] = v
        with open(filepath, 'w') as fp:
            json.dump(gen_parameters, fp)        
        

class JobGenerator(Generator):
    SERIAL = 1
    PARALLEL = 2

    def __init__(self, total_nodes, resources_types, serial_prob, parallel_prob, parallel_node_prob, performance,
                 min_request, max_request, params, max_opers_serial, max_parallel_duration, distributions=['gamma', 'expon', 'erlang', 'beta', 'arcsine']):
        """

        :param total_nodes:
        :param resources_types:
        :param serial_prob:
        :param parallel_prob:
        :param parallel_node_prob:
        :param performance:
        :param min_request:
        :param max_request:
        :param params:
        :param max_opers_serial:
        :param max_parallel_duration:
        :param distributions:
        """

        self.total_nodes = total_nodes
        self.resources = list(min_request.keys())  # resources_types
        self.performance = performance
        self.params = params['params'] if 'params' in params else {}
        self.minimal_request = min_request
        self.maximal_request = max_request
        Generator.__init__(self, distributions)
        self._init_probabilities(serial_prob, parallel_prob, parallel_node_prob)
        self.max_opers = params['max_opers'] if 'max_opers' in params else 0
        self.max_opers_serial = max_opers_serial 
        self.max_parallel_duration = max_parallel_duration

    def add_sample(self, jobs_flops, save=False):
        """

        :param jobs_flops:
        :param save:
        :return:
        """
        if not self.params:
            max_opers = max(jobs_flops)
            self.min_opers = min(jobs_flops)
            if not hasattr(self, 'max_opers') or self.max_opers < max_opers:
                self.max_opers = max_opers * 1.2  # at max 20% as in the sample
            self.params, self.density, self.bins = self._generate_dist_params(jobs_flops)
            if save:
                filename = 'job_params-{}'.format(int(time.time()))
                self._save_parameters(filename, self.params, min_opers=self.min_opers, max_opers=self.max_opers, density=self.density.tolist(), bins=self.bins.tolist())

    def next_job(self, size=1):
        """

        :param size:
        :return:
        """
        assert (self.params), 'Sample data must be added first!'
        values = []
        while len(values) < size:
            rnds = self.dist_rand(size=size - len(values), **self.params)
        
            i_rnds = searchsorted(self.bins, rnds)
            for i, rnd in zip(i_rnds, rnds): 
                if self.min_opers <= rnd <= self.max_opers:
                    if self.density[i - 1] > 0:
                        values.append(int(rnd))
                    elif random() < 0.05:
                        values.append(int(rnd))
        job_features = []
        for flops in values:
            type, nodes, request = self._generate_request(flops)
            runtime = int(self._calc_runtime(flops, nodes, request))
            job_features.append((type, runtime, nodes, request))
        return job_features

    def _central_node_prob(self, par_node_prob):
        """

        :param par_node_prob:
        :return:
        """
        central = 0
        accum_prob = 0
        for i, prob in par_node_prob.items():
            central = i
            accum_prob += prob
            if round(accum_prob) >= 0.5:
                break
        return central, accum_prob

    def _init_probabilities(self, serial_prob, parallel_prob, par_node_prob):
        """

        :param serial_prob:
        :param parallel_prob:
        :param par_node_prob:
        :return:
        """
        assert (serial_prob + parallel_prob == 1), 'Jobs types (serial, parallel) probabilities dont sum 1'
        assert (round(sum([p for p in par_node_prob.values()]), 10) in (0, 1)), 'Parallel node probabilities dont sum 1'

        _calc_node_size = lambda x: log(x) / log(2)

        self.p_serial = serial_prob
        self.p_parallel = parallel_prob
        self._nodes = [n for n in par_node_prob]
        self._node_probabilities = [p for n, p in par_node_prob.items()]
        self._node_base_min = _calc_node_size(1)
        self._node_base_max = _calc_node_size(len(self._nodes))
        central, self._nodes_prob = self._central_node_prob(par_node_prob)
        self._node_base_med = _calc_node_size(central)

    def _generate_nodes(self):
        """

        :return:
        """
        u = random()
        if u <= self._nodes_prob:
            a = self._node_base_min
            b = self._node_base_med
        else:
            a = self._node_base_med
            b = self._node_base_max
        nodes = round(2 ** self._uniform_dist(a, b))
        return nodes

    def _uniform_dist(self, a, b):
        """

        :param a:
        :param b:
        :return:
        """
        return (random() * (b - a)) + a

    def _calc_runtime(self, gflops, nodes, request):
        """

        :param gflops:
        :param nodes:
        :param request:
        :return:
        """
        part_power = sum([nodes * request[r] for r, p in self.performance.items() if r in request])
        if not part_power:
            return 0
        return gflops / part_power

    def _generate_request(self, gflops, job_profiles=None):
        """

        :param gflops:
        :param job_profiles:
        :return:
        """
        request = {k: 0 for k in self.resources}
        r = random()
        if r <= self.p_serial and gflops <= self.max_opers_serial:
            nodes = 1
            type = self.SERIAL
        else:
            nodes = self._generate_nodes()
            type = self.PARALLEL
        if job_profiles:
            self._job_request_profile(type, nodes, request)
        else:
            self._job_request(type, nodes, request)

        return type, nodes, request

    def _job_request(self, type, nodes, request):
        """

        :param type:
        :param nodes:
        :param request:
        :return:
        """
        for res in request:
            _min = self.minimal_request[res]
            _max = self.maximal_request[res]
            if type == self.PARALLEL and nodes == 1 and res == 'core':
                _min = 2
            if type == self.SERIAL:
                if not (res in ('core', 'mem')):
                    _max = 0
                if res == 'core':
                    _max = 1
            v = round(self._uniform_dist(_min, _max))
            request[res] = v

    def _job_request_profile(self, nodes, request):
        """

        :param nodes:
        :param request:
        :return:
        """
        raise NotImplementedError()


class ArriveGenerator(Generator):
    HOURS_PER_DAY = 24
    MINUTES_PER_DAY = 60 * HOURS_PER_DAY
    SECONDS_PER_DAY = 60 * MINUTES_PER_DAY

    BUCKETS = 48  # Every 30min
    SECONDS_IN_BUCKET = SECONDS_PER_DAY / BUCKETS
    DISTRIBUTIONS = ['gamma', 'dweibull', 'lognorm', 'genexpon', 'expon', 'exponnorm', 'exponweib', 'exponpow', 'truncexpon']

    def __init__(self, initial_time, hour_prob, day_prob, month_prob, param, total_jobs, cyclic_day_start=0, max_arrive_time=5 * 24 * 3600, distributions=[]):
        """

        :param initial_time:
        :param hour_prob:
        :param day_prob:
        :param month_prob:
        :param param:
        :param total_jobs:
        :param cyclic_day_start:
        :param max_arrive_time:
        :param distributions:
        """
        assert (0 <= cyclic_day_start < 24), 'The cicle must start between [0, 23].'

        self.initial_time = initial_time
        self.hour_prob = hour_prob
        self.day_prob = day_prob
        self.month_prob = month_prob
        self.cyclic_day_start = cyclic_day_start
        self.total_jobs = total_jobs
        self.initial_bucket = int(
            round((str_datetime(self.initial_time).get_hours() * self.BUCKETS) / self.HOURS_PER_DAY))
        self.distribution_plot = False

        self.time_from_begin = {}
        self.params = param
        self.current = {}
        self.weights = {}
        self.means = {}
        self.points = {}
        self.reminders = {}

        if not distributions:
            distributions = self.DISTRIBUTIONS
        Generator.__init__(self, distributions)

        self.initialized = False
        self.TOO_MUCH_ARRIVE_TIME = log(max_arrive_time)

    def add_sample(self, submission_times, rush_hours=(8, 17), save=False):
        """

        :param submission_times:
        :param rush_hours:
        :param save:
        :return:
        """
        
        total_jobs = len(submission_times)
        assert(total_jobs >= 1000), 'Data might no be representative. There are just {} jobs.'.format(total_jobs)
        _bucket_number = lambda _dtime: _dtime.get_hours() * (self.BUCKETS // self.HOURS_PER_DAY) + (
        _dtime.get_minutes() // 30)
        submission_times = sorted(submission_times)
        sample_size = len(submission_times)
        data = {type: [] for type in ['total', 'rush']}

        init, end = rush_hours

        max_arrive_time_diff = 0
        ia_times = []

        for i, (cur_time, next_time) in enumerate(zip(submission_times[:-1], submission_times[1:])):
            ia_time = abs(next_time - cur_time)  # Must be always positive... using abs just in case
            _datetime = str_datetime(cur_time)
            _hour = _datetime.get_hours()
            _pos = _bucket_number(_datetime)
            
            data['total'].append(_pos)
            ia_times.append(ia_time)
            # Store jobs that were submitted at rush hours
            if init < _hour < end:
                data['rush'].append(_pos)
                
            if ia_time > max_arrive_time_diff:
                max_arrive_time_diff = ia_time
        avg_iatimes = sum(ia_times) / len(ia_times)
        avg_percentile = percentileofscore(ia_times, avg_iatimes)
        iatimes_stdev = pstdev(ia_times)
        
        _log_arrive_time = log(max_arrive_time_diff)
        
        if self.TOO_MUCH_ARRIVE_TIME > _log_arrive_time:
            self.TOO_MUCH_ARRIVE_TIME = _log_arrive_time

        if not self.params:
            for type in data:
                if not data[type]:
                    continue
                self.params[type], _, _ = self._generate_dist_params(data[type])
    
                if self.distribution_plot:
                    self._save_distribution_plot('arrive distribution', data, type)
            if save:
                filename = 'arrive_params-{}'.format(int(time.time()))
                self._save_parameters(filename, self.params)

    def next_time(self, generation_stats):
        """

        :param generation_stats:
        :return:
        """
        assert (len(self.params) > 0), 'Data samples must be added before try to generate a next time arrive.'

        if not self.initialized:
            self._initialize()
        sample_name = 'total'
        dist_params = self.params[sample_name]
        dist = {'dist_name': dist_params['dist_name'], 'dist_param': dist_params['dist_param'], 'optional':dist_params['optional']}
        
        current_hour, current_day, current_month = self._date_info(self.time_from_begin[sample_name])
                
        current_jobs = generation_stats['current_jobs']
        togen_jobs = generation_stats['total_jobs']
        if current_jobs == 0:
            current_rate = 1
        else:
            cur = [generation_stats['current_h'][current_hour - 1] / current_jobs, generation_stats['current_d'][current_day - 1] / current_jobs, generation_stats['current_m'][current_month - 1] / current_jobs]
            prob = [self.hour_prob[current_hour - 1], self.day_prob[current_day - 1], self.month_prob[current_month - 1]]
            progress = [ ((c * current_jobs) / (togen_jobs * p)) for c, p in zip(cur, prob)]
            progress = [p for p in progress if 0 < p <= 1]
            if not progress:
                current_rate = 1
            else:
                current_rate = reduce(lambda x, y: x * y, progress)
                
        factor = (1 - current_rate)
        cur_max_ia_time = self.TOO_MUCH_ARRIVE_TIME - (self.TOO_MUCH_ARRIVE_TIME - log(self.SECONDS_IN_BUCKET)) * factor
        
        assert(cur_max_ia_time <= self.TOO_MUCH_ARRIVE_TIME), 'wrong cur max ia time {} > {} ({} days)'.format(cur_max_ia_time, self.TOO_MUCH_ARRIVE_TIME, exp(cur_max_ia_time) / (3600 * 24))
        while True:
            rnd_value = self.dist_rand(**dist)
            if rnd_value <= cur_max_ia_time: 
                break
        current_bucket = self.current[sample_name]
        weights = self.weights[sample_name]
        
        self.points[sample_name] += exp(rnd_value) / self.SECONDS_IN_BUCKET
        next_arrive = 0

        while self.points[sample_name] > weights[current_bucket]:
            self.points[sample_name] -= weights[current_bucket]
            current_bucket = (current_bucket + 1) % self.BUCKETS
            next_arrive += self.SECONDS_IN_BUCKET
        
        new_reminder = self.points[sample_name] / weights[current_bucket]
        more_time = self.SECONDS_IN_BUCKET * (new_reminder - self.reminders[sample_name])
        next_arrive += more_time
        self.reminders[sample_name] = new_reminder
        self.time_from_begin[sample_name] += int(next_arrive)

        # Update all attributes
        self.current[sample_name] = current_bucket

        return self.time_from_begin[sample_name]

    def _date_info(self, timestamp):
        dtime = str_datetime(timestamp)
        return dtime.get_hours(), dtime.get_weekday(), dtime.get_month()


    def _save_distribution_plot(self, name, data, data_type):
        """

        :param name:
        :param data:
        :param data_type:
        :return:
        """
        import matplotlib.pyplot as plt
        x = range(self.BUCKETS)
        _data = data[data_type]
        plt.hist(_data, self.BUCKETS, normed=True)
        _param = self.params[data_type]
        dist_name = _param['dist_name']
        dist_param = _param['dist_param']
        optional = _param['optional']
        dist = getattr(_statistical_distributions, dist_name)
        pdf_fitted = dist.pdf(x, *dist_param, **optional)
        plt.plot(pdf_fitted, label=dist_name)
        plt.show()

    def _initialize(self):
        """

        :return:
        """
        for _name, dist_params in self.params.items():
            dist = {'dist_name': dist_params['dist_name'], 'dist_param': dist_params['dist_param'], 'optional':dist_params['optional']}
            self.weights[_name] = [self.dist_cdf(i + 0.5, **dist) - self.dist_cdf(i - 0.5, **dist) for i
                                   in range(self.BUCKETS)]
            self.means[_name] = sum(self.weights[_name]) / self.BUCKETS
            self.weights[_name] = [w / self.means[_name] for w in self.weights[_name]]
            self.current[_name] = self.initial_bucket
            self.points[_name] = 0
            self.reminders[_name] = 0
            self.time_from_begin[_name] = self.initial_time
        self.initialized = True


class WorkloadGenerator:
    
    def __init__(self, workload, sys_config, performance, request_limits, reader_class=None, resources_target=None, user_behavior=None, walltime_calculation=None, non_processing_resources=['mem'], **kwargs):
        """

        :param workload:
        :param sys_config:
        :param performance:
        :param request_limits:
        :param reader_class:
        :param resources_target:
        :param user_behavior:
        :param walltime_calculation:
        :param non_processing_resources:
        :param kwargs:
        """
        self.walltime_calculation = walltime_calculation
        show_msg = False
        save_parameters = False
        job_parameters = {}
        arrive_parameters = {}
        job_gen_optional = {}
        arrive_gen_optional = {}
        
        config = load_config(sys_config)
        
        resources = self._set_resources(None, sys_config)
        equivalence = config.pop('equivalence', {})
        start_time = config.pop('start_time', int(time.time()))
        
        if not reader_class:
            reader_class = self._default_simple_swf_reader(workload, start_time, equivalence, resources)
               
        if not hasattr(reader_class, "next"):
            raise Exception('The reader_class must implement the next method.')
        
        if not resources_target:
            resources_target = self._set_resources(None, sys_config)
        
        total_jobs, _submissiont_times, _job_total_opers, serial_prob, nodes_parallel_prob, hour_prob, day_prob, month_prob, max_opers_serial, max_parallel_duration = self._initialize(reader_class,
                                                                                                  performance,
                                                                                                  resources,
                                                                                                  non_processing_resources)                
        if 'show_msg' in kwargs:
            show_msg = kwargs['show_msg']
        
        if 'job_parameters' in kwargs:
            job_parameters = kwargs['job_parameters']
            
        if 'job_distributions' in kwargs:
            job_gen_optional['distributions'] = kwargs['job_distributions']
            
        if 'arrive_parameters' in kwargs:
            arrive_parameters = kwargs['arrive_parameters']
            
        if 'save_parameters' in kwargs:
            save_parameters = kwargs['save_parameters']
            
        parallel_prob = 1 - serial_prob

        self.arrive_generator = ArriveGenerator(start_time, hour_prob, day_prob, month_prob, arrive_parameters, total_jobs)
        
        if show_msg:
            print('Arrive Generator samples...')
        self.arrive_generator.add_sample(_submissiont_times, save=save_parameters)
        if show_msg:
            print('Arrive Generator samples... Loaded')
            
        total_nodes = sum([d['nodes'] for d in resources.definition])
        total_resources = resources.total_resources()
        resource_types = list(resources.total_resources().keys())
        min_request, max_request = request_limits['min'], request_limits['max'] 

        self.job_generator = JobGenerator(total_nodes, resource_types, serial_prob, parallel_prob, nodes_parallel_prob,
                                           performance, min_request, max_request, job_parameters, max_opers_serial, max_parallel_duration, **job_gen_optional)

        if show_msg:
            print('Job Generator samples...')
        self.job_generator.add_sample(_job_total_opers, save=save_parameters)
        if show_msg:
            print('Job Generator samples... Loaded')

    def _initialize(self, base_reader, performance, resources, non_processing_resources):
        """

        :param base_reader:
        :param performance:
        :param resources:
        :param non_processing_resources:
        :return:
        """
        total_nodes = sum([d['nodes'] for d in resources.definition])
        resource_types = resources.total_resources().keys()

        proc_units = [res_type for res_type in resource_types if not (res_type in non_processing_resources)]

        _job_runtimes = []
        _job_submission_times = []
        _job_types = {'serial': 0, 'parallel': 0}
        _job_total_opers = SortedList()
        _job_hours = [0 for i in range(24)]
        _job_days = [0 for i in range(7)]
        _job_months = [0 for i in range(12)]
        _nodes_requested = {n + 1: 0 for n in range(total_nodes)}
        _n_parallel_requests = 0
        total_jobs = 0
        _max_opers_serial = 0
        _max_parallel_duration = 0
        while True:
            _dict = base_reader.next()
            if not _dict:
                break
            submisison_time = _dict['queued_time']
            hour, weekday, month = self._date_info(submisison_time)
            _job_hours[hour - 1] += 1
            _job_days[weekday - 1] += 1
            _job_months[month - 1] += 1
            duration = int(_dict['duration'])
            requested_nodes = int(_dict['requested_nodes'])
            requested_resources = _dict['requested_resources']

            requested_proc_units = {k: v for k, v in requested_resources.items() if k in proc_units}

            _job_runtimes.append(duration)
            _job_submission_times.append(submisison_time)
            _job_type = self._define_job_type(requested_nodes, requested_proc_units)
            _job_total_oper = self._job_operations(duration, requested_nodes, requested_proc_units, performance)
            _job_total_opers.add(_job_total_oper)

            _job_types[_job_type] += 1
            if _job_type == 'serial' and _max_opers_serial < _job_total_oper:
                _max_opers_serial = _job_total_oper
            if _job_type == 'parallel':
                if _max_parallel_duration < duration:
                    _max_parallel_duration = duration
                _nodes_requested[requested_nodes] += 1
                _n_parallel_requests += 1

            assert (requested_nodes < total_nodes), 'More nodes than available have been requested. ({} > {})'.format(
                requested_nodes, total_nodes)
            total_jobs += 1

        _job_types.update((k, v / total_jobs) for k, v in _job_types.items())
        _job_hours = [h / total_jobs for h in _job_hours]
        _job_days = [d / total_jobs for d in _job_days]
        _job_months = [m / total_jobs for m in _job_months]

        if _n_parallel_requests > 1:
            nodes_parallel_prob = {n: _nodes_requested[n] / _n_parallel_requests for n in _nodes_requested}
        
        return total_jobs, _job_submission_times, _job_total_opers, _job_types['serial'], nodes_parallel_prob, _job_hours, _job_days, _job_months, _max_opers_serial, _max_parallel_duration

    def generate_jobs(self, n, filename, overwrite=True, writer=None):
        """

        :param n:
        :param filename:
        :param overwrite:
        :param writer:
        :return:
        """
        if not writer:
            writer = DefaultWriter(filename, overwrite=overwrite)
        else:
            obj_assertion(writer, WorkloadWriter,
                          'Received {} type as system resource. resources_class type expected.',
                          [WorkloadWriter.__class__.__name__])
                
        generation_stats = {
            'current_jobs': 0,
            'total_jobs': n,
            'current_h': [0 for i in range(24)],
            'current_d': [0 for i in range(7)],
            'current_m': [0 for i in range(12)]
        }
        
        jobs = {}
        sub_times = []
        for i in range(n):
            if (i + 1) % 1000 == 0 :
                print('Generated {} jobs'.format(i + 1))
            submit_time = self.arrive_generator.next_time(generation_stats)
            sub_times.append(submit_time)
            self._update_stats(generation_stats, submit_time)
            job_type, duration, nodes, request = self.job_generator.next_job()[0]
            jobs[i + 1] = {
                'job_number': i + 1,
                'submit_time': submit_time,
                'duration': duration,
                'nodes': nodes,
                'resources': request,
                # Calculate walltime if method exists...
                'requested_time': self.walltime_calculation(i + 1, job_type, duration, nodes, request) if self.walltime_calculation else None 
            }                       
            writer.add_newline(jobs[i + 1])
        writer.close_file()
        print('Done. {} generated jobs.'.format(n))
        return jobs

    def _update_stats(self, stats, stime):
        """

        :param stats:
        :param stime:
        :return:
        """
        hour, day, month = self._date_info(stime)
        stats['current_jobs'] += 1
        stats['current_h'][hour - 1] += 1
        stats['current_d'][day - 1] += 1
        stats['current_m'][month - 1] += 1

    def _date_info(self, submission_date):
        """

        :param submission_date:
        :return:
        """
        dtime = str_datetime(submission_date)
        return dtime.get_hours(), dtime.get_weekday(), dtime.get_month()

    def _set_resources(self, resources_obj, sys_config):
        """

        :param resources_obj:
        :param sys_config:
        :return:
        """
        if resources_obj:
            obj_assertion(resources_obj, Resources,
                          'Received {} type as system resource. resources_class type expected.',
                          [resources_obj.__class__.__name__])
            return resources_obj
        elif sys_config:
            config = load_config(sys_config)
            resources_obj = Resources(node_prefix='', **config)
            return resources_obj
        else:
            raise Exception('A resources object or the path to the system config must be given.')

    def _define_job_type(self, nodes, proc_units):
        """

        :param nodes:
        :param proc_units:
        :return:
        """
        total_q = sum([v for k, v in proc_units.items()])
        if total_q == 1 and nodes == 1:
            return 'serial'
        return 'parallel'

    def _job_operations(self, duration, nodes, proc_units, performance):
        """

        :param duration:
        :param nodes:
        :param proc_units:
        :param performance:
        :return:
        """
        used_gflops = int(duration * nodes * sum([v * performance[k] for k, v in proc_units.items()]))
        return used_gflops
    
    def _update_distinct_woption(self, name, main_value, value, default_value, converter):
        """

        :param name:
        :param main_value:
        :param value:
        :param default_value:
        :param converter:
        :return:
        """
        return lambda _dict: _dict.update({name: converter(_dict.pop(main_value) if _dict[main_value] != value else _dict.pop(default_value))}) or _dict

    def _update_name(self, current, new, converter):
        """

        :param current:
        :param new:
        :param converter:
        :return:
        """
        return lambda _dict: _dict.update({new: converter(_dict.pop(current))}) or _dict

    def _default_simple_swf_reader(self, workload, start_time, equivalence, resources):
        """

        :param workload:
        :param start_time:
        :param equivalence:
        :param resources:
        :return:
        """
        reg_exp = '\s*(?P<job_id>[-+]?\d+)\s*(?P<queue_time>[-+]?\d+)\s*([-+]?\d+)\s*(?P<duration>[-+]?\d+)\s*(?P<allocated_processors>[-+]?\d+)\s*([-+]?\d+\.\d+|[-+]?\d+)\s*(?P<used_memory>[-+]?\d+)\s*(?P<requested_number_processors>[-+]?\d+)\s*(?P<expected_duration>[-+]?\d+)\s*(?P<requested_memory>[-+]?\d+)\s*([-+]?\d+)\s*(?P<user>[-+]?\d+)\s*([-+]?\d+)\s*([-+]?\d+)\s*(?P<queue_number>[-+]?\d+)\s*([-+]?\d+)\s*([-+]?\d+)\s*([-+]?\d+)'
        
        _func1 = self._update_distinct_woption('total_processors', 'requested_number_processors', '-1', 'allocated_processors', int)
        _func2 = self._update_distinct_woption('mem', 'requested_memory', '-1', 'used_memory', int)
        _func3 = self._update_name('queue_time', 'queued_time', int)
   
        tweaker = DefaultTweaker(start_time, resources, equivalence)
        
        return WorkloadFileReader(workload, reg_exp, tweaker, [_func1, _func2, _func3])



"""
MIT License

Copyright (c) 2017 AlessioNetti, cgalleguillosm

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from collections import deque
from threading import Thread, Semaphore
from multiprocessing import Process


class AsyncWriter:
    """
    This class handles asynchronous IO of files, using a thread and queue-based implementation.
    """

    def __init__(self, path, pre_process_fun=None, buffer_size=10000):
        """
        Constructor for the class

        :param path: Path to the output file
        :param pre_process_fun: A pre-processing function for objects pushed to the queue. It MUST be a function that
            receives an object as input, and returns a string representation of it
        """
        self._toTerminate = False
        self._thread = None
        self._deque = deque()
        self._sem = Semaphore(value=0)
        self._buffer_size = buffer_size
        self._buf_counter = 0
        if not pre_process_fun:
            pre_process_fun = self._dummy_pre_process
        self._pre_processor_wrapper = QueueFlusher(path, pre_process_fun)

    def push(self, data_obj):
        """
        Writes to an output file the data object specified as input asynchronously, after pre-processing it.

        :param data_obj: The object to be pre-processed to string format, and written to output
        """
        self._deque.append(data_obj)
        self._buf_counter += 1
        if self._buf_counter >= self._buffer_size:
            self._buf_counter = 0
            self._sem.release()

    def start(self):
        """
        Starts up the worker thread handling file IO
        """
        self._thread = Thread(target=self._working_loop)
        self._thread.start()

    def stop(self):
        """
        Stops the worker thread handling file IO
        """
        if self._thread is not None:
            self._toTerminate = True
            self._sem.release()
            self._thread.join()
            self._thread = None

        self._deque.clear()

    def _working_loop(self):
        while not self._toTerminate or len(self._deque) > 0:
            self._sem.acquire()
            process = Process(target=self._pre_processor_wrapper.flush, args=(self._deque, self._buffer_size))
            process.start()
            process.join()
            counter_pop = 0
            while counter_pop < self._buffer_size and len(self._deque) > 0:
                self._deque.popleft()
                counter_pop += 1

    @staticmethod
    def _dummy_pre_process(data_obj):
        return str(data_obj)

class QueueFlusher:

    def __init__(self, path, func):
        self._path = path
        self._func = func

    def flush(self, data, buffer_size, **kwargs):
        buffer = ''
        counter_flushed = 0
        while counter_flushed < buffer_size and len(data) > 0:
            entry = data.popleft()
            str_out = self._func(entry)
            if isinstance(str_out, (list, tuple)) and len(str_out) > 1:
                for str_el in str_out:
                    buffer += str_el
            else:
                buffer += str_out
            counter_flushed += 1

        with open(self._path, 'a') as f:
            f.write(buffer)



"""
MIT License

Copyright (c) 2017 cgalleguillosm, AlessioNetti

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from matplotlib.backends.backend_pdf import PdfPages
from math import floor
from accasim.utils.reader_class import DefaultReader
from accasim.utils.misc import load_config, from_isodatetime_2_timestamp as timestamp_func, str_resources
from accasim.utils.file import path_leaf, load_jsonfile
from accasim.base.resource_manager_class import Resources
from accasim.experimentation.schedule_parser import define_result_parser
from accasim.utils.misc import DEFAULT_SIMULATION
from copy import deepcopy
from os.path import splitext, join
from scipy.signal import savgol_filter
from os.path import isfile
import numpy as np


class PlotFactory:
    """
    A class for plot production and schedule files pre-processing.
    
    In this class, some basic algorithms are implemented for pre-processing the schedule files produced through 
    simulation, and for producing some common evaluation plots.
    """

    SCHEDULE_CLASS = 'schedule'
    BENCHMARK_CLASS = 'benchmark'
    SLOWDOWN_PLOT = 'slowdown'
    QUEUE_SIZE_PLOT = 'queue_size'
    LOAD_RATIO_PLOT = 'load_ratio'
    EFFICIENCY_PLOT = 'efficiency'
    SCALABILITY_PLOT = 'scalability'
    SIMULATION_TIME_PLOT = 'sim_time'
    SIMULAION_MEMORY_PLOT = 'sim_memory'

    PLOT_TYPES = {
        SCHEDULE_CLASS: [SLOWDOWN_PLOT, QUEUE_SIZE_PLOT, LOAD_RATIO_PLOT, EFFICIENCY_PLOT],
        BENCHMARK_CLASS: [SCALABILITY_PLOT, SIMULATION_TIME_PLOT, SIMULAION_MEMORY_PLOT]
    }

    def __init__(self, plot_class, sim_params_fname=None, config=None, resource=None, workload_parser=None, debug=False):
        """
        The constructor for the class.
        
        :param plot_class: the plot_class of files to be analyzed. Can be either 'schedule', if schedule files are going to be 
        analyzed, or 'benchmark' if resource usage log files will be analyzed;
        :params sim_params_fname: 
        :param config: The path to a system configuration file. Needed for the schedule meta-simulation;
        :param resource: a resource type in the system to be considered. If specified, all resource-related statistics
            will be computed in regards to this resource alone;
        :param workload_parser: 
        :param debug: Debug flag.
        """
        self._debug = debug
        if not (plot_class in self.PLOT_TYPES.keys()):
            if self._debug:
                print('Wrong Plot plot_class chosen. Selecting schedule plot_class by default...')
            plot_class = self.SCHEDULE_CLASS
        self._plot_class = plot_class

        self._sim_params_fname = sim_params_fname  # if sim_params_fname is not None and isfile(sim_params_fname) else None
        self._config = config
        self._resource = resource
        self._workload_parser = workload_parser

        self._preprocessed = False
        self._filepaths = []
        self._labels = []

        self._slowdowns = []
        self._queuesizes = []
        self._loadratiosX = []
        self._loadratiosY = []
        self._efficiencies = []

        self._simdata = []
        self._schedtimes = []
        self._mantimes = []
        self._simmemory = []
        self._scalabilitydataX = []
        self._scalabilitydataY = []

        self._resource_order = None
        if self._sim_params_fname is None:
            self._resource_order = DEFAULT_SIMULATION['RESOURCE_ORDER']

        # Base resource availability per-node (never changes)
        self._base_res = {}
        # Current resource availability per-node
        self._sys_res = {}
        # Aggregated used resources for all nodes
        self._used_res_sum = {}
        # Aggregate base resource availability for used nodes only
        self._avl_res_sum = {}
        # Aggregated base resource availability for all nodes
        self._base_res_sum = {}
        # Amount of currently used nodes
        self._used_nodes = 0
        # Number of total nodes in the system
        self._total_nodes = 0

    def set_files(self, paths, labels):
        """
        Set the paths and labels of the files to be analyzed.
        
        :param paths: A list of filepaths related to the files to be analyzed; 
        :param labels: the labels associated to each single file, used in the plots; must have the same length as paths;
        """
        self._preprocessed = False
        if not isinstance(paths, (list, tuple)):
            self._filepaths = [paths]
            self._labels = [labels]
        else:
            self._filepaths = paths
            self._labels = labels

        if len(self._filepaths) != len(self._labels):
            if self._debug:
                print("Filepaths and Labels lists must have the same lengths.")
            self._labels = []
            self._filepaths = []

    def pre_process(self, trimSlowdown=True, trimQueueSize=False):
        """
        Performs pre-processing on all specified files, according to their type.
        
        If the files are of the schedule type, a meta-simulation is run for each of them, computing data like slowdown,
        queue size, load ratios and such. If the data is of the benchmark type, the files are simply parsed and their
        information stored.
        
        :param: trimSlowdown: boolean flag. If True, slowdown values equal to 1 will be discarded. Default is True
        :param: trimQueueSize: boolean flag. If True, queue size values equal to 0 will be discarded. Default is False
        
        """
        if not self._preprocessed:
            # Perform pre-processing for schedule files
            if self._plot_class == self.SCHEDULE_CLASS:
                self._slowdowns = []
                self._queuesizes = []
                self._loadratiosX = []
                self._loadratiosY = []
                self._efficiencies = []
                self._preprocessed = True
                for f in self._filepaths:
                    # If an error is encountered on one of the files, the process is aborted
                    if not self._getScheduleData(f, self._config, self._resource, trimSlowdown, trimQueueSize):
                        self._preprocessed = False
                        break

            # Perform pre-processing for benchmark files
            elif self._plot_class == self.BENCHMARK_CLASS:
                self._simdata = []
                self._schedtimes = []
                self._mantimes = []
                self._simmemory = []
                self._scalabilitydataX = []
                self._scalabilitydataY = []
                self._preprocessed = True
                for f in self._filepaths:
                    if not self._getBenchmarkData(f):
                        self._preprocessed = False
                        break

        if not self._preprocessed:
            print("Could not process files, please ensure they are in the correct path and format.")

        return self._preprocessed

    def produce_plot(self, type, title='', scale='linear', xlim=(None, None), ylim=(None, None), legend=True, figsize=(7, 5), meansonly=False, alpha=0.005, smooth=30, output='Output.pdf', groups=1, **kwargs):
        """
        Produces a single plot on the pre-processed files.
        
        The user can produce plots among the available types. These are:
            - slowdown: a box-plot distribution plot for slowdown values across test instances
            - queue_size: a box-plot for queue size in the simulation across test instances
            - load_ratio: a distribution scatter plot for the load ratio in function of the number of used nodes, for
                test instances separately;
            - efficiency: a box-plot for resource allocation efficiency across test instances
            - scalability: a scalability plot for dispatching methods across test instances
            - sim_time: a bar plot for the simulation timings across test instances
            - sim_memory: a bar plot for memory usage across test instances
        
        :param type: the type of the plot, must be one of the above;
        :param title: the title of the plot;
        :param scale: the scale of the plot (see matplotlib documentation);
        :param xlim: the left-right bounds for axis scaling, is a tuple;
        :param ylim: the bottom-top bounds for axis scaling, is a tuple;
        :param legend: activates the legend, is a boolean;
        :param figsize: the size of the figure, is a tuple;
        :param meansonly: triggers the plot of mean values alone in box-plots, is a boolean;
        :param alpha: the alpha of certain features in plots, in particular for distribution scatter plots;
        :param smooth: smoothing factor used for the Savitzky-Golay filter in the scalabily plot. The lower the number,
            the higher the smoothing;
        :param output: path of the output PDF file;
        """
        if not self._preprocessed:
            self.pre_process()
            print("Plot_factory: Files were not pre-processed yet. Calling the pre_process method.")

        if type == self.SLOWDOWN_PLOT and self._plot_class == self.SCHEDULE_CLASS:
            self.box_plot(self._slowdowns, title=title, ylabel='Slowdown', scale=scale, xlim=xlim, ylim=ylim, figsize=figsize, meansonly=meansonly, output=output, groups=groups, **kwargs)
        elif type == self.QUEUE_SIZE_PLOT and self._plot_class == self.SCHEDULE_CLASS:
            self.box_plot(self._queuesizes, title=title, ylabel='Queue size', scale=scale, xlim=xlim, ylim=(0, None), figsize=figsize, meansonly=meansonly, output=output, groups=groups, **kwargs)
        elif type == self.LOAD_RATIO_PLOT and self._plot_class == self.SCHEDULE_CLASS:
            self.distribution_scatter_plot(self._loadratiosX, self._loadratiosY, title=title, scale=scale, xlim=(-0.01, 1.01), ylim=(-0.01, 1.01), figsize=figsize, alpha=alpha, output=output, **kwargs)
        elif type == self.EFFICIENCY_PLOT and self._plot_class == self.SCHEDULE_CLASS:
            self.box_plot(self._efficiencies, title=title, ylabel='Resource efficiency', scale=scale, xlim=xlim, ylim=ylim, figsize=figsize, meansonly=meansonly, output=output, groups=groups, **kwargs)
        elif type == self.SCALABILITY_PLOT and self._plot_class == self.BENCHMARK_CLASS:
            self.scalability_plot(self._scalabilitydataX, self._scalabilitydataY, title, scale=scale, xlim=xlim, ylim=ylim, figsize=figsize, legend=legend, smooth=smooth, output=output, **kwargs)
        elif type == self.SIMULATION_TIME_PLOT and self._plot_class == self.BENCHMARK_CLASS:
            self.box_plot_times(self._mantimes, self._schedtimes, title=title, scale=scale, xlim=xlim, ylim=ylim, figsize=figsize, legend=legend, output=output, **kwargs)
        elif type == self.SIMULAION_MEMORY_PLOT and self._plot_class == self.BENCHMARK_CLASS:
            self.box_plot_memory(self._simmemory, title=title, scale=scale, xlim=xlim, ylim=ylim, figsize=figsize, legend=legend, output=output, **kwargs)
        else:
            raise Exception("Plot type specified is not valid. Review the documentation for valid plot types.")


    def _getBenchmarkData(self, filepath):
        """
        Pre-processes a resource usage log file.
        
        :param filepath: the path to the log file;
        :return: True if successful, False otherwise;
        """
        if self._debug:
            print("- Pre-processing file " + filepath + "...")
        # Tries to read from the file, aborts if an error is encountered
        try:
            f = open(filepath)
            mantimes = []
            schedtimes = []
            mems = []
            simtime = 0
            disptime = 0
            maxqueuesize = 0

            for line in f:
                # Each line is parsed and values are extracted from it
                attrs = line.split(';')
                mantimes.append(float(attrs[4]))
                schedtimes.append((int(attrs[1]), float(attrs[3])))
                mems.append(float(attrs[5]))
                simtime += float(attrs[2])
                disptime += float(attrs[3])

                if int(attrs[1]) > maxqueuesize:
                    maxqueuesize = int(attrs[1])
            f.close()
        except Exception as e:
            raise Exception("Error encountered while pre-processing: " + str(e))

        # Certain statistics are computed from the data
        data = {}
        data['avgman'] = np.average(np.array(mantimes))
        data['avgsched'] = np.average(np.array([el[1] for el in schedtimes]))
        data['simtime'] = simtime / 1000.0
        data['schedtime'] = disptime / 1000.0
        data['mantime'] = data['simtime'] - data['schedtime']
        data['avgmem'] = np.average(np.array(mems))
        data['maxmem'] = np.max(np.array(mems))

        # The scalability data is computed through binning: we want to obtain an X, Y set, where in X are the distinct
        # queue sizes, and in Y are the average times in ms to perform dispatching on such queue sizes
        binningfactor = 1
        bins = int(floor(maxqueuesize / binningfactor))
        queuevalues = np.linspace(0, maxqueuesize, bins)
        mappinglist = []
        for i in range(bins):
            mappinglist.append([])
        step = (maxqueuesize) / (bins - 1)
        for qsize, stime in schedtimes:
            index = int(floor(qsize / step))
            mappinglist[index].append(stime)
        finallist = []
        finalqueuevalues = []
        for i in range(len(mappinglist)):
            l = mappinglist[i]
            if len(l) > 0:
                finallist.append(sum(l) / len(l))
                finalqueuevalues.append(queuevalues[i])

        self._mantimes.append(mantimes)
        self._schedtimes.append([el[1] for el in schedtimes])
        self._simmemory.append(mems)
        self._simdata.append(data)
        self._scalabilitydataX.append(finalqueuevalues)
        self._scalabilitydataY.append(finallist)
        return True

    def _getScheduleData(self, filepath, config, resource=None, trimSlowdown=True, trimQueueSize=False):
        """
        Performs pre-processing on a schedule file through a meta-simulation process.
        
        :param filepath: The path of the file to be analyzed;
        :param config: The path to the system configuration file;
        :param resource: A resource to be considered for resource-related metrics; if none is specified, all resource
            types are used;
        :param: trimSlowdown: boolean flag. If True, slowdown values equal to 1 will be discarded. Default is True
        :param: trimQueueSize: boolean flag. If True, queue size values equal to 0 will be discarded. Default is False
        :return: True if successful, False otherwise;
        """
        if self._debug:
            print("- Pre-processing file " + filepath + "...")
        # Generates the dictionary of system resources from the config file
        resobject, equiv = self._generateSystemConfig(config)
        self._base_res = resobject.availability()
        res_types = resobject.system_resource_types

        # Makes sure the resource type exists in the system
        if resource is not None and resource not in resobject.system_resource_types:
            if self._debug:
                print("Resource type " + resource + "is not valid. Using all available resources...")
            resource = None

        # Tries to read from the log file, aborts if an error is encountered
        try:
            _sim_params_path = None
            # If the simulator config path points to a file, it is considered as is
            if self._sim_params_fname is not None and isfile(self._sim_params_fname):
                _sim_params_path = self._sim_params_fname
            # If it is a plain string, it is used as a token for config files in the experimentation
            elif self._sim_params_fname is not None:
                _path, _filename = path_leaf(filepath)
                _sim_params_path = join(_path, self._sim_params_fname)
            # If it is none, the default_result_parser will use the DEFAULT_SIMULATION config

            if _sim_params_path is not None:
                _resource_order = load_jsonfile(_sim_params_path)['RESOURCE_ORDER']
            else:
                _resource_order = self._resource_order

            if self._workload_parser is not None:
                reader = DefaultReader(filepath, parser=self._workload_parser, equivalence=equiv)
            else:
                reader = DefaultReader(filepath, parser=define_result_parser(_sim_params_path), equivalence=equiv)

            slowdowns = []
            timePoints = set()
            jobs = {}
            rev_timePoints = {}
            if self._debug:
                print("Loading jobs...")
            while True:
                # Jobs are read and their slowdown values are stored
                job = reader.read()
                if job is not None:
                    job['start_time'] = timestamp_func(job['start_time'])
                    job['end_time'] = timestamp_func(job['end_time'])
                    job['queue_time'] = timestamp_func(job['queue_time'])
                    _start_time = job['start_time']
                    _end_time = job['end_time']
                    _queued_time = job['queue_time']
                    duration = _end_time - _start_time
                    wait = _start_time - _queued_time
                    slowdown = (wait + duration) / duration if duration != 0 else wait if wait != 0 else 1.0
                    if slowdown > 1.0 or not trimSlowdown:
                        slowdowns.append(slowdown)

                    job_id = job['job_id']
                    jobs[job_id] = job
                    # Timepoints for use in the simulation are stored
                    timePoints.add(_queued_time)
                    self._addToDictAsList(rev_timePoints, _queued_time, job_id, 'queue')
                    timePoints.add(_start_time)
                    self._addToDictAsList(rev_timePoints, _start_time, job_id, 'start')
                    if duration > 0:
                        timePoints.add(_end_time)
                        self._addToDictAsList(rev_timePoints, _end_time, job_id, 'end')
                else:
                    break
        except Exception as e:
            raise Exception("Error encountered while pre-processing: " + str(e))

        # It may happen that the slowdown list is empty if all jobs have a value equal to 1. In this case we add
        # a fake value, equal to 1 as well
        if trimSlowdown and len(slowdowns) == 0:
            slowdowns.append(1)

        if self._debug:
            print("Jobs loaded. Sorting...")

        # We compute the final set of distinct, ordered timepoints
        timePoints = sorted(timePoints)
        timePointsIDX = 0

        self._sys_res = deepcopy(self._base_res)
        self._base_res_sum = {k: sum(self._base_res[n][k] for n in self._base_res) for k in res_types}
        self._used_res_sum = {k: 0 for k in res_types}
        self._avl_res_sum = {k: 0 for k in res_types}
        self._used_nodes = 0
        self._total_nodes = len(self._base_res.values())

        queue = set()
        running = set()

        # Pre-allocating the lists to store performance metrics, for efficiency
        queued = [0] * len(timePoints)  # []
        resources = [0] * len(timePoints)  # []
        run = [0] * len(timePoints)  # []
        efficiency = [0] * len(timePoints)  # []
        efficiencyperjob = [0] * len(jobs)  # []
        efficiencyIDX = 0

        if self._debug:
            print("Sorting done. Starting simulation...")

        # Meta-simulation: goes on until there are no more timepoints to consider
        while timePointsIDX < len(timePoints):
            point = timePoints[timePointsIDX]
            timePointsIDX += 1

            # Adds to the queue jobs that were submitted in this timepoint
            jobstoqueue = rev_timePoints[point]['queue']
            # queue += len(jobstoqueue)
            queue.update(jobstoqueue)

            # Jobs that have terminated release their resources
            jobstoend = rev_timePoints[point]['end']
            if len(jobstoend) > 0:
                for j_id in jobstoend:
                    j = jobs[j_id]
                    req, assignations = self._getRequestedResources(_resource_order, j['assignations'])
                    self._deallocate_resources(req, assignations, resource)
                # running -= len(jobstoend)
                running = running - jobstoend

            # Jobs that have to start take their resources from the system
            jobstostart = rev_timePoints[point]['start']
            if len(jobstostart) > 0:
                for j_id in jobstostart:
                    j = jobs[j_id]
                    if j['end_time'] - j['start_time'] > 0:
                        req, assignations = self._getRequestedResources(_resource_order, j['assignations'])
                        self._allocate_resources(req, assignations, resource)
                        # running += 1
                        running.add(j_id)
                # queue -= len(jobstostart)
                queue = queue - jobstostart

                # Additionally, we store for every started job its resource allocation efficiency
                for j_id in jobstostart:
                    j = jobs[j_id]
                    if j['end_time'] - j['start_time'] > 0:
                        req, assignations = self._getRequestedResources(_resource_order, j['assignations'])
                        eff = self._getResourceEfficiency(req, assignations, self._sys_res, resource)
                        efficiencyperjob[efficiencyIDX] = eff
                        efficiencyIDX += 1

            # System metrics are computed AFTER dispatching
            queued[timePointsIDX - 1] = len(queue)  # queue
            run[timePointsIDX - 1] = len(running)  # running
            resources[timePointsIDX - 1] = self._getLoadRatio(resource)
            efficiency[timePointsIDX - 1] = self._getLoadRatioSelective(resource)

        if self._debug:
            print("Simulation done!")

        if trimQueueSize:
            queued = [q for q in queued if q != 0]
            run = [r for r in run if r != 0]

        # The metrics values for this instance are added to the internal variables
        self._slowdowns.append(slowdowns)
        self._queuesizes.append(queued)
        self._efficiencies.append(efficiencyperjob)
        self._loadratiosX.append([el[0] for el in efficiency])
        self._loadratiosY.append([el[1] for el in efficiency])
        return True

    def _addToDictAsList(self, dict, key, el, type):
        """
        Simple method that adds an element to a dictionary and creates sub-entries if needed.
        
        :param dict: The target dictionary 
        :param key: The key of the element to add
        :param el: The element to add
        :param type: The type of the element to add, used in the sub-dictionary for the key entry
        :return: None
        """
        if key not in dict:
            dict[key] = {'queue': set(), 'start': set(), 'end': set()}
        dict[key][type].add(el)

    def _allocate_resources(self, req, assignations, resource=None):
        """
        Method that allocates the resources for a certain starting job and updates all data structures related to
        resource usage
        
        :param req: The resource request of the job
        :param assignations: The list of nodes assigned to the job
        :param resource: A resource type to be considered for performance metrics (optional)
        :return: None
        """
        for node in assignations:
            # If the node goes from the unused to the used state, we update the number of used nodes and the amount
            # of available resources among the used nodes, for the efficiency plots
            if resource is None and all(self._sys_res[node][k] == self._base_res[node][k] for k in self._base_res[node].keys()):
                self._used_nodes += 1
                for k, v in self._base_res[node].items():
                    self._avl_res_sum[k] += v
            # If a specific resource type is considered, the same condition is triggered only if such resource is used
            elif resource is not None and self._sys_res[node][resource] == self._base_res[node][resource] and req[resource] > 0:
                self._used_nodes += 1
                self._avl_res_sum[resource] += self._base_res[node][resource]
            # Updating the per-node currently available resources
            for k, val in req.items():
                self._sys_res[node][k] -= val
                if self._sys_res[node][k] < 0:
                    self._sys_res[node][k] = 0
                    if self._debug:
                        print("Caution: resource " + k + " is going below zero.")

        # Updating the dictionary of per-type currently used resources
        for k, v in req.items():
            self._used_res_sum[k] += v * len(assignations)
            if self._used_res_sum[k] > self._avl_res_sum[k]:
                self._used_res_sum[k] = self._avl_res_sum[k]

    def _deallocate_resources(self, req, assignations, resource):
        """
        Method that de-allocates the resources for a certain starting job and updates all data structures related to
        resource usage
        
        :param req: The resource request of the job
        :param assignations: The list of nodes assigned to the job
        :param resource: A resource type to be considered for performance metrics (optional)
        :return: None
        """
        for node in assignations:
            for k, val in req.items():
                self._sys_res[node][k] += val
                if self._sys_res[node][k] > self._base_res[node][k]:
                    self._sys_res[node][k] = self._base_res[node][k]
                    if self._debug:
                        print("Caution: resource " + k + " is going beyond its base capacity.")
        # In this case the check for used-unused nodes must be performed after the resources are de-allocated
            if resource is None and all(self._sys_res[node][k] == self._base_res[node][k] for k in self._base_res[node].keys()):
                self._used_nodes -= 1
                for k, v in self._base_res[node].items():
                    self._avl_res_sum[k] -= v
            elif resource is not None and self._sys_res[node][resource] == self._base_res[node][resource] and req[resource] > 0:
                self._used_nodes -= 1
                self._avl_res_sum[resource] -= self._base_res[node][resource]

        # The method is specular to allocate_resources and works identically
        for k, v in req.items():
            self._used_res_sum[k] -= v * len(assignations)
            if self._used_res_sum[k] < 0:
                self._used_res_sum[k] = 0

    def _generateSystemConfig(self, config_path):
        """
        Generates a Resources object from a system configuration file.
        
        :param config_path: the path to the config file;
        :return: the Resources object and the resource equivalence;
        """
        try:
            config = load_config(config_path)
            equiv = config.pop('equivalence', {})
            # PEP 448 - Additional Unpacking Generalizations 
            # python 3.5 and newer 
            if not(node_prefix in config):
                config[node_prefix] = ''
            resources = Resources(**config)
            return resources, equiv
        except Exception as e:
            if config_path != '':
                print("Could not load system config: " + str(e))
            else:
                print("A system configuration file must be specified.")
            exit()

        return None, None

    def _getRequestedResources(self, _resource_order, assignations_str):
        """
        TO BE IMPLEMENTED:
        returns the requested resources for the input job.
        
        :param job: the dictionary related to the current job;
        :return: the dictionary of resources needed by each job unit, and the list of node assignations;
        """
        _assignations_list = assignations_str.split(str_resources.SEPARATOR)[0:-1]
        _nodes_list = [assign.split(';')[0] for assign in _assignations_list]
        _request = { k:int(v) for k, v in zip(_resource_order, _assignations_list[0].split(';')[1:])}
        return _request, _nodes_list

    def _getResourceEfficiency(self, reqres, nodes, sys_res, resource):
        """
        Computes the resource allocation efficiency metric for a certain input job.
        
        This method computed the resource allocation efficiency AFTER dispatching is performed, not before.
        
        :param reqres: the dictionary of resources requested by each job unit;
        :param nodes: the list of node assignations;
        :param sys_res: the dictionary of system resources;
        :param resource: the resource type to be considered (if present);
        :return: the resource allocation efficiency;
        """

        # Computing the amount of used resources by the job
        if resource is None:
            used = sum(r * len(nodes) for r in reqres.values())
        else:
            used = reqres[resource] * len(nodes)

        avl = 0
        # Computing the amount of available resources in nodes used by the job
        for node in set(nodes):
            if resource is None:
                avl += sum(r for r in sys_res[node].values())
            else:
                avl += sys_res[node][resource]
        return used / (avl + used)

    def _getLoadRatio(self, resource):
        """
        Returns the standard load ratio for the system.
        
        :param resource: the resource type to be considered (if present);
        :return: the load ratio;
        """
        loadratio = 0

        if resource is None:
            loadratio = sum(self._used_res_sum.values()) / sum(self._base_res_sum.values())
        elif resource in self._base_res_sum:
            loadratio = self._used_res_sum[resource] / self._base_res_sum[resource]

        return loadratio

    def _getLoadRatioSelective(self, resource):
        """
        Returns the per-step resource allocation efficiency.
        
        This is defined as a X,Y pair where X expresses the fraction of used nodes, and Y defines the fraction of used
        resources in such nodes.

        :param resource: the resource type to be considered (if present);
        :return: an X,Y pair expressing the per-step resource allocation efficiency;
        """
        loadratio = 0
        if self._used_nodes > 0:
            if resource is None:
                loadratio = sum(self._used_res_sum.values()) / sum(self._avl_res_sum.values())
            elif resource in self._avl_res_sum:
                loadratio = self._used_res_sum[resource] / self._avl_res_sum[resource]
            return self._used_nodes / self._total_nodes, loadratio
        else:
            return 0, 0

    def _getDistributionStats(self, data):
        """
        Returns some useful distribution statistics for the input data.
        
        The mean, minimum, maximum, median, and quartiles for the data are computed.
        
        :param data: The iterable for the input data;
        :return: a dictionary of statistics for the data distribution;
        """
        stats = {}
        stats['avg'] = np.average(data)
        stats['min'] = np.min(data)
        stats['max'] = np.max(data)
        stats['median'] = np.median(data)
        stats['quartiles'] = np.percentile(data, range(0, 100, 25))
        return stats

    def box_plot(self, data, title='', ylabel='', scale='linear', xlim=(None, None), ylim=(None, None), figsize=(7, 5), meansonly=False, output='Output.pdf', groups=1):
        """
        Produces a box-and-whiskers plot for the input data's distributions.
        
        :param data: the input data; must be a list, in which each element is again a list containing all of the data
            regarding a certain test instance; the ordering must be that of the labels;
        :param title: the title of the plot;
        :param ylabel: the Y-axis label;
        :param scale: the scale of the plot;
        :param xlim: the left-right axis boundaries, is a tuple;
        :param ylim: the bottom-top axis boundaries, is a tuple;
        :param figsize: the size of the figure, is a tuple;
        :param meansonly: if True only the mean values for each distribution are depicted;
        :param output: the path to the output PDF file;
        """
        cycler = ['b', 'r', 'y', 'g', 'c', 'm', 'k', 'w']
        fontsize = 12
        plt.rc('xtick', labelsize=fontsize)
        plt.rc('ytick', labelsize=fontsize)
        N = len(data)

        spacing = 0.2
        ind = [i * spacing for i in np.arange(N)]
        width = 0.1
        markersize = 250
        linecol = 'black'
        tricol = 'black'
        vertlinecol = 'gray'
        # rectcol = 'blue'

        fig, ax = plt.subplots(figsize=figsize)

        c_group = 0
        c = groups
        for i, d in enumerate(data):
            rectcol = cycler[c_group]
            mydata = self._getDistributionStats(d)
            if not meansonly:
                ax.add_patch(patches.Rectangle((ind[i], mydata.get('quartiles')[1]), width, mydata.get('quartiles')[3] - mydata.get('quartiles')[1], facecolor=rectcol, alpha=0.75))
                ax.plot([ind[i] + width / 2, ind[i] + width / 2], [mydata.get('min'), mydata.get('max')], color=vertlinecol, linestyle='-', linewidth=2, zorder=1)
                ax.scatter(ind[i] + width / 2, mydata.get('max'), marker='_', s=markersize, zorder=2, color=linecol)
                ax.scatter(ind[i] + width / 2, mydata.get('min'), marker='_', s=markersize, zorder=2, color=linecol)
                ax.scatter(ind[i] + width / 2, mydata.get('median'), marker='_', s=markersize, zorder=2, color=linecol)
                ax.scatter(ind[i] + width / 2, mydata.get('avg'), marker='^', s=markersize / 4, zorder=2, color=tricol)
            else:
                # If meansonly is True, only a path is drawn for the mean values.
                ax.add_patch(patches.Rectangle((ind[i], 0), width, mydata.get('avg'), facecolor=rectcol, alpha=0.75))
                ax.scatter(ind[i] + width / 2, mydata.get('avg'), marker='_', s=markersize / 4, zorder=0, color=linecol)
            c -= 1
            if c == 0:
                c_group += 1
                c = groups

        # add some text for labels, title and axes ticks
        ax.set_ylabel(ylabel, fontsize=fontsize)
        ax.set_xlabel('Dispatching method', fontsize=fontsize)
        ax.set_title(title)
        ax.set_xticks([i + width / 2 for i in ind])
        ax.set_xticklabels(self._labels)
        ax.set_yscale(scale)
        ax.set_ylim(top=ylim[1], bottom=ylim[0], emit=True, auto=False)
        ax.set_xlim(left=xlim[0], right=xlim[1], emit=True, auto=False)

        plt.grid(linestyle=':', color='gray', zorder=0)
        plt.show()

        ff = PdfPages(output)
        ff.savefig(fig)
        ff.close()

    def box_plot_times(self, dataman, datasched, title='', scale='linear', xlim=(None, None), ylim=(None, None), figsize=(7, 5), legend=True, output='Output.pdf'):
        """
        Produces a bar plot for the timings in the simulations, across test instances.
        
        The bars will depict the average time required to perform dispatching in each simulation step, and the
        time required to perform simulation-related tasks in the simulation.
        
        :param dataman: the data for the time required in each step to perform simulation-related tasks. Is a list,
            where each element is again a list containing the data for a certain test instance;
        :param datasched: the data for the time required in each step to perform dispatching. Is a list, where
            each element is again a list containing the data for a certain test instance;
        :param title: the title of the plot;
        :param scale: the scale of the plot;
        :param xlim: the left-right boundaries for the plot, is a tuple;
        :param ylim: the bottom-top boundaries for the plot, is a tuple;
        :param figsize: the size of the figure, is a tuple;
        :param legend: enables or disables visualization of the legend;
        :param output: the path to the output file;
        """

        fontsize = 12
        plt.rc('xtick', labelsize=fontsize)
        plt.rc('ytick', labelsize=fontsize)
        N = len(dataman)

        spacing = 0.2
        ind = [i * spacing for i in np.arange(N)]
        width = 0.1
        markersize = 250

        fig, ax = plt.subplots(figsize=figsize)

        for i in range(N):
            avgman = np.average(np.array(dataman[i]))
            avgsched = np.average(np.array(datasched[i]))
            if i == 0:
                ax.add_patch(patches.Rectangle((ind[i], 0), width, avgman, facecolor='orange', edgecolor='black', hatch='//', alpha=0.75))  # , label='Simulation'))
                ax.add_patch(patches.Rectangle((ind[i], avgman), width, avgsched, facecolor='blue', edgecolor='black', hatch='\\', alpha=0.75, label='Dispatching decision'))
            else:
                ax.add_patch(patches.Rectangle((ind[i], 0), width, avgman, facecolor='orange', edgecolor='black', hatch='//', alpha=0.75))
                ax.add_patch(patches.Rectangle((ind[i], avgman), width, avgsched, facecolor='blue', edgecolor='black', hatch='\\', alpha=0.75))
            ax.scatter(ind[i] + width / 2, avgman + avgsched, marker='_', s=markersize / 4, zorder=0, color='black')

        # add some text for labels, title and axes ticks
        ax.set_ylabel('Time [ms]', fontsize=fontsize)
        ax.set_xlabel('Dispatching method', fontsize=fontsize)
        ax.set_title(title)
        ax.set_xticks([i + width / 2 for i in ind])
        if legend:
            ax.legend()
        ax.set_xticklabels(self._labels)
        ax.set_yscale(scale)
        ax.set_ylim(top=ylim[1], bottom=ylim[0], emit=True, auto=False)
        ax.set_xlim(left=xlim[0], right=xlim[1], emit=True, auto=False)

        plt.grid(linestyle=':', color='gray', zorder=0)
        plt.setp(plt.gca().get_legend().get_texts(), fontsize=fontsize)
        plt.show()

        ff = PdfPages(output)
        ff.savefig(fig)
        ff.close()

    def box_plot_memory(self, data, title='', scale='linear', xlim=(None, None), ylim=(None, None), figsize=(7, 5), legend=True, output='Output.pdf'):
        """
        Produces a bar plot for the memory usage in the simulations, across test instances.
        
        The bars depict average and maximum memory usage in the simulation.

        :param data: the data for memory usage in each simulation step. Is a list, where
            each element is again a list containing the data for a certain test instance;
        :param title: the title of the plot;
        :param scale: the scale of the plot;
        :param xlim: the left-right boundaries for the plot, is a tuple;
        :param ylim: the bottom-top boundaries for the plot, is a tuple;
        :param figsize: the size of the figure, is a tuple;
        :param legend: enables or disables visualization of the legend;
        :param output: the path to the output file;
        """
        fontsize = 12
        plt.rc('xtick', labelsize=fontsize)
        plt.rc('ytick', labelsize=fontsize)
        N = len(data)

        spacing = 0.2
        ind = [i * spacing for i in np.arange(N)]
        width = 0.1
        markersize = 250

        fig, ax = plt.subplots(figsize=figsize)

        for i in range(N):
            avgmem = np.average(np.array(data[i]))
            maxmem = np.max(np.array(data[i]))
            if i == 0:
                ax.add_patch(patches.Rectangle((ind[i], 0), width, avgmem, facecolor='orange', edgecolor='black', hatch='//', alpha=0.75, label='Avg. Mem'))
                ax.add_patch(patches.Rectangle((ind[i], avgmem), width, maxmem - avgmem, facecolor='blue', edgecolor='black', hatch='\\', alpha=0.75, label='Max. Mem'))
            else:
                ax.add_patch(patches.Rectangle((ind[i], 0), width, avgmem, facecolor='orange', edgecolor='black', hatch='//', alpha=0.75))
                ax.add_patch(patches.Rectangle((ind[i], avgmem), width, maxmem - avgmem, facecolor='blue', edgecolor='black', hatch='\\', alpha=0.75))
            ax.scatter(ind[i] + width / 2, maxmem, marker='_', s=markersize / 4, zorder=0, color='black')

        ax.set_ylabel('Average Memory Usage [MB]', fontsize=fontsize)
        ax.set_xlabel('Dispatching method', fontsize=fontsize)
        ax.set_title(title)
        ax.set_xticks([i + width / 2 for i in ind])
        if legend:
            ax.legend()
        ax.set_xticklabels(self._labels)
        ax.set_yscale(scale)
        ax.set_ylim(top=ylim[1], bottom=ylim[0], emit=True, auto=False)
        ax.set_xlim(left=xlim[0], right=xlim[1], emit=True, auto=False)

        plt.grid(linestyle=':', color='gray', zorder=0)
        plt.setp(plt.gca().get_legend().get_texts(), fontsize=fontsize)
        plt.show()

        ff = PdfPages(output)
        ff.savefig(fig)
        ff.close()

    def scalability_plot(self, xdata, ydata, title='', scale='linear', xlim=(None, None), ylim=(None, None), figsize=(7, 5), legend=True, smooth=30, linestyles=None, markers=None, output='Output.pdf'):
        """
        Creates a scalability plot for all test instances, where X represents the queue size, and Y the average
        time required by each dispatching method in the instances.
        
        :param xdata: the X data, containing the queue sizes for each test instance; is a list, where each element
            contains a list with the data for each test instance;
        :param ydata: the Y data, containing the average times required to perform dispatching in each test instance;
            is a list, where each element contains a list with the data for each test instance;
        :param title: the title of the plot;
        :param scale: the scale of the plot;
        :param xlim: the left-right boundaries for the plot, is a tuple;
        :param ylim: the bottom-top boundaries for the plot, is a tuple;
        :param figsize: the size of the figure, is a tuple;
        :param legend: enables or disables visualization of the legend; 
        :param smooth: smoothing factor for the Savitzky-Golay filter. The lower the number, the higher the smoothing;
        :param output: the path of the output file;
        """

        fontsize = 12
        plt.rc('xtick', labelsize=fontsize)
        plt.rc('ytick', labelsize=fontsize)
        if not linestyles:
            linestyles = ('-', '-', '--', '--', '-.', '-.', ':', ':')
        if not markers:
            markers = (None, 'o', None, '^', None, 's', None, 'p')
        
        numstyles = len(linestyles)

        fig, ax = plt.subplots(figsize=figsize)

        divideFactor = smooth

        for i in range(len(xdata)):
            markeroffset = floor(max(xdata[i]) / 20 + i * 2)
            if divideFactor > 1 and len(ydata[i]) >= divideFactor:
                win_len = floor(len(ydata[i]) / divideFactor)
                win_len += (win_len + 1) % 2
                if win_len < 5:
                    win_len = 5
                yfiltered = savgol_filter(ydata[i], win_len, 3)
            else:
                yfiltered = ydata[i]
            ax.plot(xdata[i], yfiltered, label=self._labels[i], linestyle=linestyles[i % numstyles], marker=markers[i % numstyles], markevery=markeroffset, zorder=2 if markers[i % numstyles] is None else 0)

        ax.set_ylabel('Time [ms]', fontsize=fontsize)
        ax.set_xlabel('Queue size', fontsize=fontsize)
        ax.set_title(title)
        if legend:
            ax.legend()

        ax.set_yscale(scale)
        ax.set_ylim(top=ylim[1], bottom=ylim[0], emit=True, auto=False)
        ax.set_xlim(left=xlim[0], right=xlim[1], emit=True, auto=False)

        plt.grid(linestyle=':', color='gray', zorder=0)
        plt.setp(plt.gca().get_legend().get_texts(), fontsize=fontsize)
        plt.show()

        ff = PdfPages(output)
        ff.savefig(fig)
        ff.close()

    def distribution_scatter_plot(self, xdata, ydata, title='', scale='linear', xlim=(0, 1.05), ylim=(0, 1.05), figsize=(7, 5), alpha=0.005, output='Output.pdf'):
        """
        Creates a distribution scatter plot for the system's resource efficiency. 
        
        The X values represent the amount of used nodes in a certain time step, while the Y values represent the
        fraction of used resources in such nodes. Darker areas of the plot represent values with higher frequency.
        The method creates one plot per test instance, automatically.
        
        :param xdata: 
        :param ydata: 
        :param alpha: the alpha to be used for each dot in the plot;
        :param title: the title of the plot;
        :param scale: the scale of the plot;
        :param xlim: the left-right boundaries for the plot, is a tuple;
        :param ylim: the bottom-top boundaries for the plot, is a tuple;
        :param figsize: the size of the figure, is a tuple;
        :param output: the path to the output files: the label for each test instance will be automatically added 
            for each file;
        """

        for i in range(len(xdata)):
            fig, ax = plt.subplots(figsize=figsize)

            ax.scatter(xdata[i], ydata[i], color='black', alpha=alpha, s=5)

            ax.set_title(title)
            ax.set_xlabel('Used Nodes')
            ax.set_ylabel('Used Resources')
            ax.set_yscale(scale)
            ax.set_ylim(top=ylim[1], bottom=ylim[0], emit=True, auto=False)
            ax.set_xlim(left=xlim[0], right=xlim[1], emit=True, auto=False)
            ax.grid(True)

            plt.show()
            splitoutput = splitext(output)
            ff = PdfPages(splitoutput[0] + '-' + self._labels[i] + '.pdf')
            ff.savefig(fig)
            ff.close()

    def get_preprocessed_benchmark_data(self):
        """
        Returns all of the pre-processed benchmark-related data.
        
        A tuple is returned; each element of the tuple is related to a specific kind of metric that was processed.
        Also, each element of the tuple is a list, with as many entries as the files that were processed, in the
        same order. Each element of these lists contains then the data related to a specific metric, for a specific
        test instance. All data is stored in standard Python lists.

        :return: a tuple in which every element is a list containing, in each element, a specific kind of data 
            regarding one of the test instances. The tuple contains, in this order:
            
            - the resource usage statistics' dictionaries;
            - the lists of dispatching times for each time step;
            - the lists of management times for each time step;
            - the lists of memory usage values for each time step;
            - the X scalability data containing the queue size for each test instance;
            - the Y scalability data containing the average dispatching times for each test instance;
        """
        if not self._preprocessed or self._plot_class != self.BENCHMARK_CLASS:
            return None, None, None, None, None, None
        else:
            return self._simdata, self._schedtimes, self._mantimes, self._simmemory, self._scalabilitydataX, self._scalabilitydataY

    def get_preprocessed_schedule_data(self):
        """
        Returns all of the pre-processed schedule-related data.

        A tuple is returned; each element of the tuple is related to a specific kind of metric that was processed.
        Also, each element of the tuple is a list, with as many entries as the files that were processed, in the
        same order. Each element of these lists contains then the data related to a specific metric, for a specific
        test instance. All data is stored in standard Python lists.

        :return: a tuple in which every element is a list containing, in each element, the data regarding one of the
            test instances. The tuple contains, in this order:

            - the slowdown values for jobs;
            - the queue sizes for all time steps;
            - the resource allocation efficiencies for all jobs;
            - the X data regarding the load ratios (fraction of used nodes) for all time steps;
            - the Y data regarding the load ratios (fraction of used resources) for all time steps;
        """
        if not self._preprocessed or self._plot_class != self.SCHEDULE_CLASS:
            return None, None, None, None, None
        else:
            return self._slowdowns, self._queuesizes, self._efficiencies, self._loadratiosX, self._loadratiosY

if __name__ == '__main__':
    # This is an example. It should not be executed here, but in a script in the project's root, where also
    # basic_example.py is, so that all imports can be resolved correctly.
    resultpath = ['Path/to/benchmark/file',
                  'Path/to/benchmark/file2']
    resultlabel = ['Label',
                   'Label2']
    plots = PlotFactory('benchmark')
    plots.set_files(resultpath, resultlabel)
    plots.pre_process()
    plots.produce_plot(type='scalability', title='My Scalability Plot')
